# -*- coding: utf-8 -*-
"""

nuka.utils.secret
========================

.. autoclass:: secret

nuka.utils.chmod
========================

.. autofunction:: chmod

nuka.utils.chown
========================

.. autofunction:: chown

nuka.utils.makedirs
========================

.. autofunction:: makedirs

"""
import io
import os
import sys
import time
import string
import logging
import threading
import subprocess


_write_lock = threading.Lock()

PY3 = sys.version_info[0] == 3

if not PY3:
    def next(o):
        return o.next()
_next = next


try:
    import zlib
except ImportError:
    zlib = None

try:
    import ujson as json
except ImportError:
    import json  # NOQA

try:
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve  # NOQA

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO  # NOQA

try:
    import importlib
except ImportError:
    importlib = None  # NOQA


LOG = 60
logging.addLevelName(LOG, 'LOG')
CHANGED = logging.WARNING + 1
logging.addLevelName(CHANGED, 'CHANGED')
PROGRESS = logging.WARNING + 2
logging.addLevelName(PROGRESS, 'PROGRESS')


def default_watcher(delay=5):
    def watcher(task, process):
        start = time.time()
        inc = delay
        while True:
            if task.is_alive(process):
                value = time.time() - start
                if value > inc:
                    inc += delay
                    task.send_progress(
                        'is running for {0}s'.format(int(value)),
                        level=logging.DEBUG)
                yield
            else:
                # process is dead
                yield
    return watcher


def import_module(name):
    if importlib is not None:
        return importlib.import_module(name)
    return __import__(name, globals(), locals(), [''])


def makedirs(dirname, mod=None, own=None):
    """create directories. return ``{'changed': True|False}``"""
    changed = False
    fut = ''
    dirname = dirname.rstrip('/')
    for p in dirname.split('/'):
        fut += p + '/'
        if not os.path.isdir(fut):
            os.makedirs(fut)
            changed = True
            if mod:
                chmod(fut, mod)
            if own:
                chown(fut, own)
    return dict(changed=changed)


def chmod(dst, mod, recursive=False):
    """chmod using command line"""
    if not os.path.exists(dst):
        raise OSError('{0} does not exist'.format(dst))
    if isinstance(mod, int):
        if recursive:
            raise RuntimeError()
        os.chmod(dst, mod)
    else:
        cmd = ['chmod']
        if recursive:
            cmd.append('-R')
        if isinstance(mod, (list, tuple)):
            mod = '{0}:{1}'.format(*mod)
        cmd.extend([mod, dst])
        subprocess.check_call(cmd,
                              stdin=subprocess.PIPE,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE)


def chown(dst, own, recursive=False):
    """chown using command line"""
    if not os.path.exists(dst):
        raise OSError('{0} does not exist'.format(dst))
    cmd = ['chown']
    if recursive:
        cmd.append('-R')
    if isinstance(own, (list, tuple)):
        own = '{0}:{1}'.format(*own)
    cmd.extend([own, dst])
    subprocess.check_call(cmd,
                          stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE)


def best_executable():
    executables = [
        '/usr/bin/python3.6',
        '/usr/bin/python3.5',
        '/usr/bin/python3.4',
        '/usr/bin/python2.7',
    ]
    for executable in executables:
        if os.path.isfile(executable):
            return executable
    return sys.executable


def isexecutable(path):
    return os.access(path, os.X_OK)


class secret(object):
    """secret word generation::

        >>> s = secret('something')
        >>> s.next()
        'F)>x|o;J7sOhWV~F'
        >>> s.next()
        'DP@:?|v%LaSB2v?b'

    You can use your own alphabet::

        >>> s = secret('something', alphabet='abc')
        >>> s.next()
        'bbbbcaaacacbbcbb'

    Or just ascii letters/digits::

        >>> s = secret('something', alphabet='ascii')
        >>> s.next()
        'FxoJ7sOhWVFYRN7v'

    You can also change the length::

        >>> s = secret('something', alphabet='ascii', length=3)
        >>> s.next()
        'Fxo'
    """

    def __init__(self, value, alphabet=None, length=16):
        self.value = value
        if alphabet == 'ascii':
            valids = [ord(a) for a in string.ascii_letters]
            valids += [ord(a) for a in string.digits]
        elif alphabet is None:
            valids = list(range(33, 127))
        else:
            valids = [ord(a) for a in alphabet]
        self.valids = [n for n in valids if n not in (34, 35, 39, 61, 92)]
        self.length = length
        self.iterator = self.iterator()

    def iterator(self):
        import hashlib
        i = 0
        while True:
            pw = ''
            j = 0
            i += 1
            while len(pw) < self.length:
                j += 1
                v = '{0}-{1}-{2}'.format(self.value, i, j)
                h = hashlib.sha512(v.encode('utf8'))
                pw += ''.join(chr(c) for c in h.digest() if c in self.valids)
            yield pw[:self.length]

    def next(self):
        return next(self.iterator)


def proto_dumps(data, content_type=u'plain'):
    """json.dumps() with headers. py2/3 compat"""
    data = json.dumps(data)
    if not isinstance(data, bytes):
        data = data.encode('utf8')
    if content_type == u'zlib':
        if zlib is not None:
            data = zlib.compress(data)
        else:
            content_type = u'plain'
    headers = (
        u'Content-type: {0}\nContent-Length: {1}\n'
    ).format(content_type, len(data)).encode('utf8')
    return headers + data


def proto_dumps_std(data, std, content_type='plain'):
    """json.dumps() to std with headers. py2/3 compat"""
    data = proto_dumps(data, content_type=content_type)
    std = getattr(std, 'buffer', std)
    std.write(data)


def proto_dumps_std_threadsafe(data, std):
    _write_lock.acquire()
    content_type = zlib is None and u'plain' or u'zlib'
    try:
        proto_dumps_std(data, std, content_type=content_type)
        std.flush()
    finally:
        _write_lock.release()


def proto_loads_std(std):
    """json.loads() from std with headers. py2/3 compat"""
    if isinstance(std, bytes):
        std = io.BytesIO(std)
        std.seek(0)
    else:
        std = getattr(std, 'buffer', std)
    content_type = std.readline()
    if isinstance(content_type, bytes):
        content_type = content_type.decode('utf8')
    try:
        content_type = content_type.split(':')[1].strip()
    except IndexError:
        raise ValueError(content_type)
    content_length = std.readline()
    if isinstance(content_length, bytes):
        content_length = content_length.decode('utf8')
    try:
        content_length = int(content_length.split(':')[1].strip())
    except IndexError:
        raise ValueError(content_length)
    data = std.read(content_length)
    if content_type == 'zlib':
        data = zlib.decompress(data)
    if isinstance(data, bytes):
        data = data.decode('utf8')
    try:
        data = json.loads(data)
    except ValueError:
        raise ValueError(data)
    return data



# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals
import os
import sys
import time
import codecs
import signal
import select
import difflib
import logging
import subprocess

from nuka import utils


def safe_iterator(iterator=None):
    while True:
        if iterator is not None:
            try:
                utils._next(iterator)
            except:
                logging.exception(iterator)
        yield


class RemoteTask(object):

    stdin = getattr(sys.stdin, 'buffer', sys.stdin)
    logfile = utils.StringIO()

    ignore_errors = False

    def __getattr__(self, attr):
        if attr.startswith('_'):
            raise AttributeError(attr)
        try:
            return self.res[attr]
        except KeyError:
            raise AttributeError(attr)


class Task(RemoteTask):
    """Remote task"""

    current_process_watcher = None
    current_process = None
    current_cmd = None

    process_watcher_delay = 2

    alarm_delay = 1
    alarm_count = 0

    remote_start = time.time()
    remote_calls = []

    stds = dict(
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    def __init__(self, **kwargs):
        self.start = time.time()
        self.res = {}
        self.args = kwargs

    @classmethod
    def texts_diff(self, old_text, new_text, **kwargs):
        old_list = old_text.splitlines(1)
        new_list = new_text.splitlines(1)
        res = self.lists_diff(old_list, new_list, **kwargs)
        return res

    @classmethod
    def lists_diff(self, old_list, new_list,
                   fromfile='(old)', tofile='(new)', **kwargs):
        res = difflib.unified_diff(
            old_list, new_list, fromfile=fromfile, tofile=tofile, **kwargs)
        return ''.join(res)

    @classmethod
    def get_inventory(self):
        dirname = os.path.dirname(sys.argv[0])
        cache = os.path.join(dirname, 'inventory.json')
        data = self.get()
        with codecs.open(cache) as fd:
            utils.json.read(data, fd)
        return data

    @property
    def is_debian(self):
        return os.path.isfile('/etc/debian_version')

    @property
    def is_centos(self):
        return os.path.isfile('/etc/centos-release')

    def do(self):
        """called on the remote host.
        Task arguments are available in self.args
        Must return a dict like::

            {'rc': 0, 'changed': False}
        """
        raise NotImplementedError()

    @classmethod
    def sh(self, args, stdin=b'', shell=False, env=None, check=True,
           watcher=None, short_args=None, stdout=None, stderr=None):
        """run a shell command"""
        start = time.time()
        env_ = os.environ.copy()
        env_.update(LC_ALL='C', LANG='C')
        if env:
            env_.update(env)
        kwargs = dict(self.stds, shell=shell)
        if stdout is not None:
            kwargs['stdout'] = stdout
        if stderr is not None:
            kwargs['stderr'] = stderr
        res = {'stdout': '', 'stderr': ''}
        try:
            p = subprocess.Popen(args, env=env_, **kwargs)
        except Exception:
            rc = res['rc'] = 1
            exc = self.format_exception()
        else:
            self.current_cmd = short_args or args
            self.current_process = p
            if watcher is not None:
                self.current_process_watcher = safe_iterator(watcher(self, p))
            else:
                self.current_process_watcher = safe_iterator()
            if stdin and not isinstance(stdin, bytes):
                stdin = stdin.encode('utf8')
            stdout, stderr = p.communicate(stdin)
            self.current_process = None
            utils._next(self.current_process_watcher)
            self.current_process_watcher = None
            if stdout is None:
                stdout = kwargs['stdout']
                with open(stdout.name) as fd:
                    try:
                        stdout = fd.read()
                    except OSError:
                        stdout = ''
            if isinstance(stdout, bytes):
                stdout = stdout.decode('utf8')
            if stderr is None:
                stderr = kwargs['stderr']
                with open(stderr.name) as fd:
                    try:
                        stderr = fd.read()
                    except OSError:
                        stderr = ''
            if isinstance(stderr, bytes):
                stderr = stderr.decode('utf8')
            rc = p.returncode
            res.update(rc=rc, stdout=stdout, stderr=stderr)
            exc = None
        t = time.time() - start
        self.remote_calls.append(dict(res, cmd=args, start=start, time=t,
                                      exc=exc))
        if check:
            return self.check(res)
        return res

    @classmethod
    def format_exception(self):
        import traceback
        exc_type, exc_value, exc_traceback = sys.exc_info()
        return traceback.format_exception(exc_type, exc_value,
                                          exc_traceback)

    @classmethod
    def on_alarm(self, *args):
        self.alarm_count += 1
        ready = select.select([sys.stdin], [], [], .0)[0]
        if ready:
            try:
                data = utils.proto_loads_std(self.stdin)
            except ValueError:
                pass
            else:
                sig = data.get('signal')
                if sig is not None:
                    self.on_sigint()
        if self.alarm_count % self.process_watcher_delay == 0:
            if self.current_process_watcher is not None:
                utils._next(self.current_process_watcher)
        signal.alarm(self.alarm_delay)

    @classmethod
    def on_sigint(self, *args):
        res = dict(rc=1, signal='SIGINT', current_process=None)
        if self.current_process is not None:
            # forward signal to current process if any
            p = self.current_process
            p.send_signal(signal.SIGINT)
            # let the watcher know that we no longer have a process
            if self.current_process_watcher is not None:
                self.current_process = None
                utils._next(self.current_process_watcher)
            p.wait()
            res['current_process'] = {
                'cmd': self.current_cmd,
                'pid': p.pid,
                'rc': p.returncode}
        # clean exit
        self.exit(res)

    @classmethod
    def is_alive(self, process):
        """return True iif the process is the current process"""
        if self.current_process is not None:
            return self.current_process.pid == process.pid

    @classmethod
    def send_progress(self, value, level=utils.PROGRESS):
        """send progression of the current remote process to the client"""
        if self.current_process is not None:
            p = self.current_process
            if p.returncode is None:
                message = 'sh({cmd}, pid={pid}) {value}'.format(**{
                    'cmd': ' '.join(self.current_cmd),
                    'pid': p.pid,
                    'value': value,
                    })
                self.send_log(message, level=level)

    @classmethod
    def send_log(self, message, level=logging.DEBUG):
        """send a log line to the client"""
        self.send_message(dict(message_type='log', msg=message, level=level))

    @classmethod
    def send_message(self, message):
        if 'message_type' not in message:
            raise ValueError(
                'No message_type specified in {0}'.format(message))
        utils.proto_dumps_std_threadsafe(message, sys.stdout)
        sys.stdout.flush()

    @classmethod
    def exit(self, res):
        # tel the watcher that the process is ended
        if self.current_process_watcher is not None:
            self.current_process = None
            utils._next(self.current_process_watcher)
        res.setdefault('rc', 0)
        res.setdefault('message_type', 'exit')
        res.setdefault('signal', None)
        res['meta'] = dict(
            remote_calls=self.remote_calls,
            remote_time=time.time() - self.remote_start
        )
        self.logfile.seek(0)
        logs = self.logfile.read()
        if logs.strip():
            res['log'] = logs
        if 'diff' in res:
            res.setdefault('changed', bool(res['diff']))
        utils.proto_dumps_std_threadsafe(res, sys.stdout)
        sys.exit(0)

    @classmethod
    def check(self, res):
        if res.get('rc') != 0:
            self.exit(res)
        return res

    @classmethod
    def from_dict(cls, data):
        args = data['args']
        ob = cls(**args)
        return ob



# -*- coding: utf-8 -*-
from nuka.tasks import yum
import pytest
import os

pytestmark = [
    pytest.mark.skipif(
        'centos' not in os.environ['ENV_NAME'], reason='centos only'),
]


@pytest.mark.asyncio
async def test_update_doc(host):
    res = await yum.update(cache=3600)
    assert bool(res)


@pytest.mark.asyncio
async def test_install_doc(host):
    res = await yum.install(['python'])
    assert bool(res)


@pytest.mark.asyncio
async def test_install_diff(host, diff_mode):
    with diff_mode:
        res = await yum.install(packages=['moreutils'])
        assert res.rc == 0
        assert '+moreutils\n' in res.res['diff'], res.res['diff']

        res = await yum.install(packages=['python'])
        assert res.rc == 0
        assert '+python\n' not in res.res['diff'], res.res['diff']



# -*- coding: utf-8 -*-
# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import signal
import logging
import tempfile

from nuka.task import Task
from nuka import utils


def main(data=None):

    if data is None:
        try:
            data = utils.proto_loads_std(Task.stdin)
        except Exception:
            res = dict(rc=1, exc=Task.format_exception())
            Task.exit(res)
        if 'environ' in data:
            os.environ.update(data['environ'])
        tempfile.tempdir = data['remote_tmp']

        logging.basicConfig(
            format='%(levelname)s:%(message)s',
            level=data.pop('log_level'),
            stream=Task.logfile)

    res = {}

    module, klass_name = data['task']

    try:
        mod = utils.import_module(module)
        task = getattr(mod, klass_name).from_dict(data)
    except Exception:
        res = dict(rc=1, exc=Task.format_exception())
        Task.exit(res)

    # clean exit on SIGINT
    signal.signal(signal.SIGALRM, task.on_alarm)
    signal.alarm(task.alarm_delay)

    if data['diff_mode']:
        meth_name = 'diff'
    else:
        meth_name = 'do'

    res = {}

    try:
        meth = getattr(task, meth_name)
    except AttributeError:
        res = dict(rc=0, stderr='{0} not supported'.format(meth_name))
        task.exit(res)

    try:
        res = meth() or {}
    except Exception:
        res = dict(rc=1, exc=task.format_exception())
    task.exit(res)


def setup():

    logging.basicConfig(
        format='%(levelname)s:%(message)s',
        level=logging.WARN,
        stream=Task.logfile)

    # check if we have a venv
    python = os.path.join(
        os.path.abspath(os.path.dirname(sys.argv[0])),
        'bin', 'python')
    same_dir = os.path.dirname(sys.executable) == os.path.dirname(python)
    if os.path.isfile(python) and not same_dir:
        os.execv(python, [python] + sys.argv)
    elif not os.path.isfile(python):
        # get best binary
        from nuka.utils import best_executable
        python = best_executable()
        if os.path.isfile(python) and sys.executable != python:
            os.execv(python, [python] + sys.argv)

    # launch setup task
    main(data=dict(
        task=('nuka.tasks.setup', 'setup'),
        args={},
        diff_mode=False,
        log_level=logging.WARNING)
    )


if __name__ == '__main__':
    if '--setup' in sys.argv:
        setup()
    else:
        main()



# -*- coding: utf-8 -*-
import os
import pytest
import asyncio
import logging

from nuka.tasks.user import (create_user, delete_user)
from nuka.task import wait_for_boot
from nuka.hosts import DockerContainer
from nuka.hosts import Vagrant
import nuka

nuka.config['log']['dirname'] = '.nuka/logs'
nuka.config['log']['levels'] = {
    'stream_level': logging.DEBUG,
    'file_level': logging.DEBUG,
    'remote_level': logging.DEBUG,
}


@pytest.yield_fixture(scope='session')
def session_container(request):
    loop = asyncio.get_event_loop()
    ENV_NAME = os.environ['ENV_NAME']
    if ENV_NAME.endswith('vagrant'):
        h = Vagrant()
    else:
        PY, IMAGE, TAG = ENV_NAME.split('-', 2)
        if IMAGE == 'nukai':
            IMAGE = 'bearstech/nukai'
        DOCKER_IMAGE = '{0}:{1}'.format(IMAGE, TAG)
        DOCKER_NAME = DOCKER_IMAGE.replace(':', '-').replace('/', '-')
        h = DockerContainer(
            hostname=DOCKER_NAME,
            image=DOCKER_IMAGE,
            command=['bash', '-c', 'while true; do sleep 1000000000000; done'],
        )
        loop.run_until_complete(wait_for_boot(h))
    # check if we can use coverage
    res = loop.run_until_complete(
        h.run_command('ls {remote_dir}/bin/coverage'.format(**nuka.config)))
    if res['rc'] == 0 and 'vagrant' not in ENV_NAME:
        h.vars['coverage'] = '{remote_dir}/bin/coverage'.format(**nuka.config)

    yield h

    # FIXME: get coverage results
    if 'coverage' in h.vars and 'vagrant' not in ENV_NAME:
        loop = asyncio.new_event_loop()
        h.loop = loop
        h._cancelled = False
        loop.run_until_complete(
            h.run_command('{coverage} combine; true'.format(**h.vars),
                          wait=False))
        res = loop.run_until_complete(
            h.run_command('cat .coverage', wait=False))
        data = res['stdout'].decode('utf8')
        for script in ('script.py', 'plugin.py'):
            data = data.replace(
                os.path.join(nuka.config['remote_dir'], 'nuka', script),
                os.path.join(os.path.dirname(nuka.__file__),
                             'remote', script)
            )
        data = data.replace(
            os.path.join(nuka.config['remote_dir'], 'nuka'),
            os.path.dirname(nuka.__file__))
        filename = os.environ['COVERAGE_REMOTE_FILE']
        if os.path.isfile(filename):
            filename += '.1'
        with open(filename, 'wb') as fd:
            fd.write(data.encode('utf8'))


@pytest.yield_fixture(scope='function')
def host(request, session_container, event_loop):
    session_container._cancelled = False
    session_container.log.warning(
        '============ START {0} =============='.format(request.function))
    session_container.loop = event_loop
    yield session_container
    session_container.log.warning(
        '============ END {0} =============='.format(request.function))


@pytest.yield_fixture(scope='function')
def user(request, host):
    host._cancelled = False
    try:
        host.loop.run_until_complete(asyncio.wait_for(
            delete_user('test_user', host=host),
            timeout=5))
    except RuntimeError:
        pass
    host.loop.run_until_complete(asyncio.wait_for(
        create_user(username='test_user', host=host),
        timeout=20))
    yield 'test_user'
    host._cancelled = False
    try:
        host.loop.run_until_complete(asyncio.wait_for(
            delete_user('test_user', host=host),
            timeout=5))
    except RuntimeError:
        pass


@pytest.fixture(scope='session')
def wait(request):
    def wait(coro, timeout=.5):
        return asyncio.wait_for(coro, timeout)
    return wait


class _diff_mode:

    def __enter__(self, *args, **kwargs):
        nuka.cli.args.diff = True

    def __exit__(self, *args, **kwargs):
        nuka.cli.args.diff = False

    def __call__(self, value):
        nuka.cli.args.diff = value


@pytest.yield_fixture(scope='function')
def diff_mode(request):
    yield _diff_mode()
    nuka.cli.args.diff = False



#!../bin/python
import nuka
from nuka.hosts import DockerContainer
from nuka.tasks import apt


async def install(host):
    res = await apt.update()
    print(res.res)
    res = await apt.install(['procps', 'varnish', 'nginx'])
    print(res.res)


nuka.run(install(DockerContainer('apt')))



# -*- coding: utf-8 -*-
"""

nuka.run
========

.. autofunction:: run

nuka.wait
==========

.. autoclass:: wait

nuka.Event
===========

.. autoclass:: Event

nuka.cancel_on_error
====================

.. autofunction:: cancel_on_error
"""
import concurrent
import functools
import logging
import asyncio
import atexit
import signal
import shutil
import sys
import os

import uvloop

# api
from nuka.cli import cli
from nuka import reports
from nuka import process
from nuka import utils  # NOQA / API
from nuka.task import wait  # NOQA / API
from nuka.task import teardown
from nuka.configuration import config

sleep = asyncio.sleep  # NOQA / API


run_vars = {}


class Event(asyncio.Future):
    """A named event that you can wait:

    .. code-block:: python

        event = Event('myevent')

        async def do_something(host):
            nuka.wait(event)
    """

    def __init__(self, name, loop=None):
        self.name = str(name)
        super().__init__(loop=loop)
        self.res = {}

    def set_result(self, **kwargs):
        """update Event.res' dict with kwargs and release the event"""
        self.res.update(kwargs)
        super().set_result(kwargs)

    def release(self):
        """release the event"""
        self.set_result()

    def __repr__(self):
        s = super().__repr__()
        s = s.replace('<Event ', '<Event {0} '.format(self.name))
        return s


def cancel_on_error(*futures):
    """Cancel futures when the coroutine raise:

    .. code-block:: python

        @cancel_on_error(event)
        async def do_something(host):
            event.release()
    """
    def wrapper(func):
        @functools.wraps(func)
        async def do(host, *args, **kwargs):
            try:
                return await func(host, *args, **kwargs)
            except asyncio.CancelledError:
                for f in futures:
                    if not f.done():
                        f.cancel()
                raise
            except:
                host.log.exception(func)
                for f in futures:
                    if not f.done():
                        f.cancel()
        return do
    return wrapper


def run(*coros, timeout=None):
    """Run coroutines:

    .. code-block:: python

        nuka.run(do_something(host)
    """
    # parse args if not already done
    if cli.args is None:
        cli.parse_args()

    # register signal if not already done
    if 'sigint' not in run_vars:
        run_vars['sigint'] = 0
        loop.add_signal_handler(signal.SIGINT, on_sigint)

    to_run = []

    for coro in coros:
        host = None
        try:
            host = coro.cr_frame.f_locals.get('host')
        except:
            pass
        if host is not None and not host.failed():
            host.log('{0}({1})'.format(coro.__name__, host))
        to_run.append((host, coro))
    coro = asyncio.gather(*[c for h, c in to_run],
                          loop=loop, return_exceptions=True)
    coro = asyncio.wait_for(coro, loop=loop, timeout=timeout)
    try:
        results = loop.run_until_complete(coro)
    except Exception as e:
        raise asyncio.CancelledError()
    else:
        res_with_exc = []
        for (host, coro), res in zip(to_run, results):
            if isinstance(res, asyncio.CancelledError):
                if host is not None:
                    if host.failed():
                        res = host.failed()
                if not isinstance(res, LookupError):
                    sys.exit(1)
            elif isinstance(res, Exception):
                raise res
            res_with_exc.append(res)
        return res_with_exc


def run_all(coro, *hosts, **kwargs):
    return run(*[coro(h) for h in hosts], **kwargs)


def on_sigint(*args, **kwargs):
    hosts = config['all_hosts'].values()
    run_vars['sigint'] = run_vars['sigint'] + 1
    if run_vars['sigint'] == 1:
        process.close_awaiting_tasks()
        logging.warning('Exiting...')
        all_tasks = 0
        for host in hosts:
            tasks = host.running_tasks()
            if not host.fully_booted.done():
                host.log.info('Cancelling {0}...'.format(tasks))
                executor.shutdown(wait=False)
            else:
                all_tasks += len(tasks)
                if tasks:
                    host.log.info('Waiting for {0}...'.format(tasks))
            # do not use host.cancel() because we don't want to cancel running
            # tasks
            host._cancelled = True
    elif run_vars['sigint'] == 2:
        logging.warning('Killing remote processes...')
        for host in hosts:
            loop.create_task(host.send_messages(dict(signal='SIGINT')))
    elif run_vars['sigint'] > 2:
        logging.warning('Exploding...')
        executor.shutdown(wait=False)
        sys.exit(1)


def on_exit():
    if cli.finalized and not cli.help:
        if 'all_hosts' in config and 'remote_dir' in config:
            hosts = config['all_hosts'].values()
            hosts = [h for h in hosts if h._tasks]
            coros = [teardown(host=h) for h in hosts if h.loop is loop]
            if coros:
                loop.run_until_complete(asyncio.wait(coros))
            if hosts:
                reports.build_reports(hosts)
        executor.shutdown(wait=True)
        process.close_connections()
        loop.close()
        dirname = config['tmp']
        if os.path.isdir(dirname):
            shutil.rmtree(dirname)
    if 'exit_message' in run_vars:
        print(run_vars['exit_message'])


# do no use cli to parse this option sinc we need the loop early
if '--uvloop' in sys.argv:
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    loop = asyncio.get_event_loop()
    level = logging.WARNING
elif '--debug' in sys.argv:
    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    level = logging.DEBUG
else:
    loop = asyncio.get_event_loop()
    level = logging.WARNING

logging.basicConfig(stream=sys.stdout, level=level,
                    format='%(levelname)-5.5s: %(message)s')

# explicit executor that we can shutdown gracefully
executor = concurrent.futures.ThreadPoolExecutor(5)
loop.set_default_executor(executor)


if 'TESTING' not in os.environ:
    # do nothing on exit while unittests
    atexit.register(on_exit)



# -*- coding: utf-8 -*-
import os
from nuka import config

config['inventory_modules'].append('nuka.inventory.net')
config['templates'].append(os.path.join(
        os.path.dirname(__file__), 'templates'))



"""
yum related tasks
"""
import os
import time
import codecs

from nuka.task import Task


class update(Task):
    """yum update"""

    timestamp_file = '/root/.last-yum-update'

    def __init__(self, cache=None, **kwargs):
        kwargs.setdefault('name', '')
        kwargs.update(cache=cache)
        super(update, self).__init__(**kwargs)

    def do(self):
        cache = self.args['cache']
        timestamp_file = self.args.get('timestamp_file', self.timestamp_file)
        if cache:
            try:
                mtime = os.path.getmtime(timestamp_file)
            except OSError:
                need_update = True
            else:
                need_update = time.time() < mtime + cache
        else:
            need_update = True

        if need_update:
            res = self.sh(['yum', 'update', '-q', '-y'])
            if cache:
                with codecs.open(timestamp_file, 'w', 'utf8') as fd:
                    fd.write(str(time.time()))
            res['changed'] = True
        else:
            res = dict(rc=0, changed=False)
        return res


class install(Task):
    """yum install"""

    def __init__(self, packages=None, update_cache=None, **kwargs):
        kwargs.setdefault('name', ', '.join(packages or []))
        kwargs.update(packages=packages, update_cache=update_cache)
        super(install, self).__init__(**kwargs)

    def pkg_list(self, packages):
        installed = []
        res = self.sh(['yum', '-q', 'list', 'installed'] + (packages or []),
                      check=False)
        in_list = False
        for line in res['stdout'].split('\n'):
            if not line:
                continue
            elif line.lower().startswith('installed packages'):
                in_list = True
                continue
            elif not in_list:
                continue
            line = line.split()
            package = line[0]
            package = package.split('.', 1)[0]
            installed.append(package)
        return installed

    def do(self):
        """install packages"""
        packages = self.args['packages']
        if not packages:
            return dict(rc=1, stderr='no packages provided')
        installed = self.pkg_list(packages)
        to_install = [p for p in packages if p not in installed]
        if to_install:
            update_cache = self.args.get('update_cache')
            if update_cache is not None:
                update(cache=update_cache).do()
            res = self.sh(['yum', 'install', '-q', '-y'] + packages)
        else:
            res = dict(rc=0)
        res['changed'] = to_install
        return res

    def diff(self):
        packages = self.args['packages']
        installed = self.pkg_list(packages)
        installed = [p + '\n' for p in sorted(installed)]
        packages = [p + '\n' for p in sorted(packages)]
        diff = self.lists_diff(installed, packages)
        return dict(rc=0, diff=diff, packages=packages)



# -*- coding: utf-8 -*-
from nuka.remote.task import Task


def test_remote_task():
    p = Task.from_dict(dict(name='yo', args={}))

    res = p.sh(['ls'], env=dict(K='V'))
    assert res['rc'] == 0
    assert len(res['stdout'])
    assert not len(res['stderr'])

    res = p.sh(['ls', '/nope'], env=dict(K='V'), check=False)
    assert res['rc'] == 2
    assert not len(res['stdout'])
    assert len(res['stderr'])

    res = p.sh(['lsss', '/nope'], check=False)
    assert res['rc'] == 1



#!../bin/python
from nuka.hosts import DockerCompose
from nuka.tasks import shell
import nuka

hosts = DockerCompose(project_name='myproject')
nuka.run(hosts.boot())

host = hosts['myproject_debian_stretch_1']


async def my_tasks(host):
    await shell.shell('whoami')

nuka.run(my_tasks(host))



# -*- coding: utf-8 -*-
from nuka.tasks import file
import pytest
import os

pytestmark = [
    pytest.mark.skipif(
        'gawel' not in os.getenv('GPG_AGENT_INFO', ''),
        reason='gawel not found in GPG_AGENT_INFO')
]


@pytest.mark.asyncio
async def test_gpg_file(host):
    res = await file.put([dict(src='tests/templates/gpg.txt.gpg',
                               dst='/tmp/gpg.txt')])
    assert bool(res)
    res = await file.cat('/tmp/gpg.txt')
    assert res.content == 'yo {{name}}\n'


@pytest.mark.asyncio
async def test_gpg_template(host):
    res = await file.put([dict(src='gpg.j2.gpg', dst='/tmp/gpg.j2')],
                         ctx=dict(name='dude'))
    assert bool(res)
    res = await file.cat('/tmp/gpg.j2')
    assert res.content == 'yo dude\n'



# -*- coding: utf-8 -*-
from nuka.tasks import apt
from nuka.tasks import git
import pytest
import os

pytestmark = [
    pytest.mark.skipif(
        'debian' not in os.environ['ENV_NAME'], reason='debian only'),
]


@pytest.mark.asyncio
async def test_clone(host):
    assert (await apt.install(packages=['git']))
    # FIXME: branch sucks on wheezy. maybe because an old git version
    # res = await git.git(
    #     src='https://github.com/gawel/aiocron.git',
    #     dst='/tmp/nuka_clone',
    # )
    # assert res
    res = await git.git(
        src='https://github.com/gawel/aiocron.git',
        dst='/tmp/nuka_clone',
        tag='0.1',
    )
    res = await git.git(
        src='https://github.com/gawel/aiocron.git',
        dst='/tmp/nuka_clone',
        tag='0.6',
    )
    assert res
    with pytest.raises(RuntimeError):
        await git.git(
            src='https://github.com/gawel/aiocron.git',
            dst='/tmp/nuka_clone',
            tag='0.6', branch='master'
        )



# -*- coding: utf-8 -*-
from nuka.tasks import shell
import pytest


@pytest.mark.asyncio
async def test_sh(host):
    assert (await shell.shell('ls / | grep etc'))


@pytest.mark.asyncio
async def test_command(host):
    assert (await shell.command(['ls', '/']))


@pytest.mark.asyncio
async def test_commands(host):
    assert (await shell.commands([['ls', '/']]))



# -*- coding: utf-8 -*-
import nuka


def test_event():
    e = nuka.Event('one')
    e.set_result(rc=1)
    assert e.res['rc'] == 1

    e = nuka.Event('two')
    e.release()
    assert e.done()

    e = nuka.Event('three')
    assert '<Event three' in repr(e)



# -*- coding: utf-8 -*-
"""
Host
====

.. autoclass:: HostGroup
   :members:

.. autoclass:: Host
   :members:

Docker
======

.. autoclass:: DockerCompose
   :members:

.. autoclass:: DockerContainer
   :members:

Vagrant
=======

.. autoclass:: Vagrant
   :members:


Libcloud
========

.. autoclass:: Cloud
   :members:
"""
from .base import all_hosts  # NOQA
from .base import Host  # NOQA
from .base import LocalHost  # NOQA
from .base import Chroot  # NOQA
from .base import HostGroup  # NOQA
from .vagrant import Vagrant  # NOQA

try:
    import compose  # NOQA
except ImportError:  # pragma: no cover
    # docker is not available
    pass
else:
    from .docker_host import DockerContainer  # NOQA
    from .docker_host import DockerCompose  # NOQA

try:
    import libcloud  # NOQA
    import novaclient # NOQA
except ImportError:  # pragma: no cover
    # licloud is not available
    pass
else:
    from .cloud import Cloud  # NOQA
    from .cloud import get_cloud  # NOQA
    from .cloud import Provider  # NOQA



# -*- coding: utf-8 -*-
import nuka
from nuka.tasks import apt
from nuka.tasks import mysql
from nuka.tasks import file
from nuka.tasks import service


mysql_ready = nuka.Event('mysql_ready')


async def install_mysql(host, mysql_password):
    await apt.install(
            debconf={'mysql-server': mysql_password},
            packages=['mysql-server'], update_cache=3600)
    await service.start('mysql')
    await mysql.my_cnf(password=mysql_password)


@nuka.cancel_on_error(mysql_ready)
async def install_master(
        host, mysql_password,
        db_name=None, db_user=None, db_password=None,
        slave=None):
    await install_mysql(host, mysql_password)
    conf = await file.update(
        dst='/etc/mysql/my.cnf',
        replaces=[
            ('^bind-address.+=.+\S+$',
             'bind-address = ' + host.private_ip),
            ('^\#server-id', 'server-id'),
            ('^\#*log_bin', 'log_bin'),
            ('^expire_logs_days.+=.+\S+$', 'expire_logs_days = 1'),
            ('^\#*binlog_do_db.+=.+\S+$', 'binlog_do_db = wp'),
        ],
        appends=[
            ('^binlog_do_db =', 'binlog_format = ROW'),
            ('^binlog_format =', 'transaction-isolation = READ-COMMITTED'),
        ])
    if conf.changed:
        await service.restart('mysql')
    if db_name:
        await mysql.create_db(name=db_name, user=db_user, password=db_password)
    if slave is not None:
        await mysql.execute(
            name='GRANT REPLICATION TO {0}'.format(slave),
            sql='''
                GRANT REPLICATION SLAVE ON *.*
                TO 'replica'@'{0}' IDENTIFIED BY '{1}';
                FLUSH PRIVILEGES;'''.format(slave.private_ip, mysql_password))
    mysql_ready.release()


async def install_slave(host, mysql_password):
    await install_mysql(host, mysql_password)
    conf = await file.update(
        dst='/etc/mysql/my.cnf',
        replaces=[
            ('^\#server-id.+=.+$', 'server-id = 2'),
        ],
        appends=[
            ('^server-id', 'relay_log_space_limit = 5G'),
            ('^relay_log_space_limit', 'replicate-do-db = wp'),
        ])
    if conf.changed:
        await nuka.wait(mysql_ready)
        await service.restart('mysql')



#!/usr/bin/env python3.5
import nuka
from nuka.hosts import DockerContainer
from nuka.tasks import (shell, file)

# setup a docker container using the default image
host = DockerContainer('mycontainer')


async def do_something(host):

    # we just echoing something using the shell.command task
    await shell.command(['echo', 'it works'], host=host)

    # if no host is provided, then a var named `host` is searched
    # from the stack. Mean that this will works to
    await shell.command(['echo', 'it works too'])


async def do_something_else(host):

    # log /etc/resolv.conf content
    res = await file.cat('/etc/resolv.conf')
    host.log.info(res.content)


# those coroutines will run in parallell
nuka.run(
    do_something(host),
    do_something_else(host),
)



# -*- coding: utf-8 -*-
import asyncio
import pytest
from nuka.hosts import base


def test_basehost():
    host = base.BaseHost(address='127.0.0.1')
    assert host.hostname == '127.0.0.1'
    assert host.bootstrap_command is None

    host = base.BaseHost(address='127.0.0.1', bootstrap_command='ls')
    assert host.bootstrap_command == 'ls'

    assert len(host.running_tasks()) == 0


def test_host():
    host = base.Host(hostname='localhost')
    assert 'ls' in host.wraps_command_line('ls')[-1]

    assert 'su' in host.wraps_command_line('ls', switch_user='user')[-1]
    host.use_sudo = True
    assert 'sudo' in host.wraps_command_line('ls', switch_user='user')[-1]
    assert 'sudo' in host.wraps_command_line('ls', switch_user='root')[-1]

    assert 'user' in host.wraps_command_line('ls',
                                             switch_ssh_user='user')


def test_localhost():
    host = base.LocalHost()
    assert host.wraps_command_line('ls') == ['bash', '-c', 'ls']


@pytest.mark.asyncio
async def test_host_session(host):
    loop = host.loop
    host = base.BaseHost(hostname='localhost')
    host.loop = loop

    assert len(host._sessions) == 0
    await host.acquire_session_slot()
    assert len(host._sessions) == 1
    host.free_session_slot()
    assert len(host._sessions) == 0


@pytest.mark.asyncio
async def test_host_cancelled(host):
    host.cancel()
    with pytest.raises(asyncio.CancelledError):
        await host.run_command('ls')


@pytest.mark.asyncio
async def test_run_command(host, wait):
    res = await host.run_command('ls /')
    assert b'etc\n' in res['stdout']

    res = await host.run_command('cat -', stdin=b'echo', close_stdin=True)
    assert b'echo' in res['stdout']

    res = await host.run_command('ls /nope')
    assert res['rc'] == 2



# -*- coding: utf-8 -*-
"""
apt related tasks
"""
import os
import time
import codecs

from nuka.tasks import http
from nuka.task import Task

import logging as log

GPG_HEADER = b'-----BEGIN PGP PUBLIC KEY BLOCK-----'


def apt_watcher(delay, fd):
    """watcher for apt using APT::Status-Fd"""
    def watcher(task, process):
        start = time.time()
        inc = delay
        new_line = last_sent = None
        while True:
            if task.is_alive(process):
                value = time.time() - start
                if value > inc:
                    line = fd.readline()
                    while line:
                        if line.startswith(('dlstatus:', 'pmstatus:')):
                            line = line.strip()
                            new_line = line
                        line = fd.readline()
                    if new_line != last_sent:
                        last_sent = new_line
                        inc += delay
                        task.send_progress(new_line.split(':', 3)[-1])
                yield
            else:
                # process is dead
                yield
    return watcher


class source(Task):
    """add an apt source"""

    def __init__(self, name=None, src=None, key=None, update=True, **kwargs):
        super(source, self).__init__(name=name, src=src, key=key,
                                     update=update, **kwargs)

    def add_key(self, key):
        if isinstance(key, str):
            if key.startswith('http'):
                res = http.fetch(src=key).do()
                dst = res['dst']
            else:
                dst = key
            with open(dst, 'rb') as fd:
                data = fd.read()
            fname = '/etc/apt/trusted.gpg.d/{0}.gpg'.format(self.args['name'])
            if GPG_HEADER in data:
                self.sh('gpg --dearmor > {0}'.format(fname),
                        shell=True, stdin=data)
            else:
                with open(fname, 'wb') as fd:
                    fd.write(data)
        elif isinstance(key, tuple):
            keyserver, keyid = key
            self.sh([
                'apt-key', 'adv',
                '--keyserver', keyserver,
                '--recv-keys', keyid])

    def do(self):
        name = self.args['name']
        src = self.args['src'].strip()
        src += '\n'
        dst = os.path.join('/etc/apt/sources.list.d', name + '.list')
        changed = True
        if os.path.isfile(dst):
            with codecs.open(dst, 'r', 'utf8') as fd:
                if fd.read() == src:
                    changed = False
        if changed:
            key = self.args['key']
            if key is not None:
                self.add_key(key)
            with codecs.open(dst, 'w', 'utf8') as fd:
                fd.write(src)
            if self.args['update']:
                cmd = [
                    'apt-get', 'update',
                    '-oDir::Etc::sourcelist=' + dst,
                    '-oDir::Etc::sourceparts=-',
                    '-oAPT::Get::List-Cleanup=0'
                ]
                self.sh(cmd)
        return dict(rc=0, changed=changed)

    def diff(self):
        name = self.args['name']
        src = self.args['src'].strip()
        src += '\n'
        dst = os.path.join('/etc/apt/sources.list.d', name + '.list')
        if os.path.isfile(dst):
            with codecs.open(dst, 'r', 'utf8') as fd:
                old_data = fd.read()
        else:
            old_data = ''
        if old_data != src:
            diff = self.texts_diff(old_data, src, fromfile=dst)
        else:
            diff = u''
        return dict(rc=0, diff=diff)


class update(Task):
    """apt get update"""

    timestamp_file = '/root/.last-apt-get-update'

    def __init__(self, cache=None, **kwargs):
        kwargs.setdefault('name', '')
        kwargs.update(cache=cache)
        super(update, self).__init__(**kwargs)

    def do(self):
        cache = self.args['cache']
        timestamp_file = self.args.get('timestamp_file', self.timestamp_file)
        if cache:
            try:
                mtime = os.path.getmtime(timestamp_file)
            except OSError:
                need_update = True
            else:
                need_update = time.time() > mtime + cache
        else:
            need_update = True

        if need_update:
            kwargs = {}
            args = ['apt-get', '--force-yes', '-y', '--fix-missing']
            watch = self.args.get('watch')
            if watch:
                r, w = os.pipe2(os.O_NONBLOCK)
                kwargs['stdout'] = os.fdopen(w)
                kwargs['watcher'] = apt_watcher(watch, os.fdopen(r))
                kwargs['short_args'] = ['apt-get', 'update']
                args.extend(['-oAPT::Status-Fd=1', 'update'])
                res = self.sh(args, **kwargs)
                res['stdout'] = ''
            else:
                res = self.sh(args + ['update'], **kwargs)
            if cache:
                with codecs.open(timestamp_file, 'w', 'utf8') as fd:
                    fd.write(str(time.time()))
            res['changed'] = True
        else:
            res = dict(rc=0, changed=False)
        return res


class list(Task):

    ignore_errors = True

    def __init__(self, update_cache=None, **kwargs):
        kwargs.update(update_cache=update_cache)
        super(list, self).__init__(**kwargs)

    def do(self):
        update_cache = self.args.get('update_cache')
        if update_cache is not None:
            res = update(cache=update_cache).do()

        res = self.sh(['apt-get', 'upgrade', '-qq', '-s'], check=False)

        return res


class search(Task):

    def __init__(self, packages, **kwargs):
        kwargs.setdefault('name', ', '.join(packages or []))
        kwargs.update(packages=packages)
        super(search, self).__init__(**kwargs)

    def do(self):
        query = self.sh(['dpkg-query',
                         '-f', "'${Package}#${Status}#${Version}~\n'",
                         '-W'] + self.args['packages'],
                        check=False)
        if query['rc'] == 1:
            # not an error for dpkg-query
            query['rc'] = 0
        return query


class upgrade(Task):

    ignore_errors = True

    def __init__(self, packages=None, debconf=None,
                 debian_frontend='noninteractive', **kwargs):
        kwargs.setdefault('name', ', '.join(packages or []))
        kwargs.update(packages=packages, debconf=debconf,
                      debian_frontend=debian_frontend)
        super(upgrade, self).__init__(**kwargs)

    def do(self):
        self.sh(['rm', '/var/lib/apt/lists/partial/*'], check=False)
        env = {}
        for k in ('debian_priority', 'debian_frontend'):
            v = self.args.get(k)
            if v:
                env[k.upper()] = v
        kwargs = {'env': env}
        #  no specific package :
        if not self.args['packages']:
            res = self.sh([
                'apt-get', '-qq', '-y',
                '-oDpkg::Options::=--force-confdef',
                '-oDpkg::Options::=--force-confold', 'upgrade'
                ], check=False, **kwargs)
            return res
        else:
            to_upgrade = []
            miss_packages = []
            #  we check for all package it they are endeed installed
            for package in self.args['packages']:
                is_present = self.sh(['dpkg-query',
                                      '-f', '\'${Status}\'',
                                      '-W', package],
                                     check=False)
                log.warn(is_present['stdout'])
                if is_present['rc'] or \
                   " installed" not in is_present['stdout']:
                    #  we don't want installed package
                    miss_packages.append(package)
                    continue
                to_upgrade.append(package)
            if to_upgrade:
                cmd = ['apt-get', '-qq', '-y',
                       '-oDpkg::Options::=--force-confdef',
                       '-oDpkg::Options::=--force-confold',
                       '--only-upgrade', 'install'
                       ] + to_upgrade
                res = self.sh(cmd, check=False, **kwargs)
            else:
                res = dict(rc=0, stdout='no upgrade')
                res['changed'] = False
            res['miss_packages'] = miss_packages
            res['packages'] = to_upgrade
            return res


class debconf_set_selections(Task):
    """debconf-set-selections"""

    diff = False

    def __init__(self, selections=None, **kwargs):
        super(debconf_set_selections, self).__init__(
            selections=selections, **kwargs)

    def do(self):
        selections = []
        for selection in self.args['selections']:
            selections.append(' '.join(selection))
        res = self.sh('debconf-set-selections',
                      stdin='\n'.join(selections), check=True)
        return res


class install(Task):
    """apt get install"""

    debconf = {
        'mysql-server': (
            ['mysql-server/root_password', 'password'],
            ['mysql-server/root_password_again', 'password'],
        ),
    }

    def __init__(self, packages=None, debconf=None,
                 debian_frontend='noninteractive', debian_priority=None,
                 update_cache=None, install_recommends=False, **kwargs):
        kwargs.setdefault('name', ', '.join(packages or []))
        kwargs.update(packages=packages, debconf=debconf,
                      debian_priority=debian_priority,
                      debian_frontend=debian_frontend,
                      update_cache=update_cache,
                      install_recommends=install_recommends,
                      )
        super(install, self).__init__(**kwargs)

    def get_packages_list(self, packages):
        splited = dict([(p.split('/', 1)[0], p) for p in packages])
        cmd = ['apt-cache', 'policy'] + [k for k in splited.keys()]
        res = self.sh(cmd, check=False)
        package = source = None
        packages = {}
        for line in res['stdout'].split('\n'):
            sline = line.strip()
            if not line.startswith(' '):
                if package:
                    packages[package['name']] = package
                package = {'name': line.strip()[:-1]}
                source = None
            elif sline.startswith(('Installed:', 'Candidate:')):
                key, value = sline.split(':', 1)
                value = value.strip()
                if value.lower() == '(none)':
                    value = False
                package[key.lower()] = value
            elif sline.startswith('***'):
                source = sline.split()[-1] + ' '
                if source.startswith('0'):
                    package['source'] = splited[package['name']]
                    source = None
            elif source and sline.startswith(source):
                package['source'] = sline
                source = None
        installed = []
        for name, fullname in splited.items():
            package = packages.get(name, {})
            if name in packages:
                if package.get('installed'):
                    if '/' in fullname:
                        name, source = fullname.split('/', 1)
                        if source in package.get('source', ''):
                            installed.append(fullname)
                    else:
                        installed.append(fullname)
        return installed

    def do(self):
        """install packages"""
        packages = self.args['packages']
        debconf = self.args['debconf']
        if not packages:
            return dict(rc=1, stderr='no packages provided')
        installed = self.get_packages_list(packages)
        to_install = [p for p in packages if p not in installed]
        if to_install:
            watch = self.args.get('watch')
            update_cache = self.args.get('update_cache')
            if update_cache is not None:
                update(cache=update_cache, watch=watch).do()
            if debconf:
                for p in to_install:
                    conf = debconf.get(p, [])
                    for i, c in enumerate(self.debconf.get(p, [])):
                        if isinstance(conf, list):
                            v = conf[i]
                        else:
                            v = conf
                        stdin = ' '.join([p] + c + [v])
                        self.sh(['debconf-set-selections'], stdin=stdin)
            env = {}
            for k in ('debian_priority', 'debian_frontend'):
                v = self.args.get(k)
                if v:
                    env[k.upper()] = v
            kwargs = {'env': env}
            args = ['apt-get', 'install', '-qqy',
                    '-oDpkg::Options::=--force-confold']
            if not self.args.get('install_recommends'):
                args.append('--no-install-recommends')
            if watch:
                r, w = os.pipe2(os.O_NONBLOCK)
                kwargs['stdout'] = os.fdopen(w)
                kwargs['watcher'] = apt_watcher(watch, os.fdopen(r))
                kwargs['short_args'] = ['apt-get', 'install']
                args.extend(['-oAPT::Status-Fd=1'] + packages)
                res = self.sh(args, **kwargs)
                res['stdout'] = ''
            else:
                res = self.sh(args + packages, **kwargs)
        else:
            res = dict(rc=0)
        res['changed'] = to_install
        return res

    def diff(self):
        packages = self.args['packages']
        installed = self.get_packages_list(packages)
        to_install = [p for p in packages if p not in installed]
        installed = [p + '\n' for p in sorted(set(installed))]
        packages = [p + '\n' for p in sorted(set(packages))]
        diff = self.lists_diff(installed, packages)
        return dict(rc=0, diff=diff, changed=to_install)



# -*- coding: utf-8 -*-
import pytest
from nuka import reports
from nuka.tasks.shell import command


@pytest.mark.asyncio
async def test_reports(host):
    await command(['ls'])
    reports.build_reports([host])



import os
import codecs
import socket
import ctypes
import ctypes.util
from ctypes import (
    Structure, Union, POINTER,
    pointer, get_errno, cast,
    c_ushort, c_byte, c_void_p, c_char_p, c_uint, c_uint16, c_uint32
)


libc = ctypes.CDLL(ctypes.util.find_library('c'))


class struct_sockaddr(Structure):
    _fields_ = [
        ('sa_family', c_ushort),
        ('sa_data', c_byte * 14)
    ]


class struct_sockaddr_in(Structure):
    _fields_ = [
        ('sin_family', c_ushort),
        ('sin_port', c_uint16),
        ('sin_addr', c_byte * 4)
    ]


class struct_sockaddr_in6(Structure):
    _fields_ = [
        ('sin6_family', c_ushort),
        ('sin6_port', c_uint16),
        ('sin6_flowinfo', c_uint32),
        ('sin6_addr', c_byte * 16),
        ('sin6_scope_id', c_uint32)
    ]


class union_ifa_ifu(Union):
    _fields_ = [
        ('ifu_broadaddr', POINTER(struct_sockaddr)),
        ('ifu_dstaddr', POINTER(struct_sockaddr))
    ]


class struct_ifaddrs(Structure):
    pass


struct_ifaddrs._fields_ = [
    ('ifa_next', POINTER(struct_ifaddrs)),
    ('ifa_name', c_char_p),
    ('ifa_flags', c_uint),
    ('ifa_addr', POINTER(struct_sockaddr)),
    ('ifa_netmask', POINTER(struct_sockaddr)),
    ('ifa_ifu', union_ifa_ifu),
    ('ifa_data', c_void_p)
]


def iter_ifaps(ifap):
    ifa = ifap.contents
    if ifa:
        yield ifa
        while ifa.ifa_next:
            ifa = ifa.ifa_next.contents
            yield ifa


def getinfos(ifa):
    try:
        sa = ifa.ifa_addr.contents
    except ValueError:
        raise StopIteration()
    family = sa.sa_family
    addr = None
    if family == socket.AF_INET6:
        sa = cast(pointer(sa), POINTER(struct_sockaddr_in6)).contents
        addr = socket.inet_ntop(family, sa.sin6_addr)
        yield family, {'address': addr}
    elif family == socket.AF_INET:
        sa = cast(pointer(sa), POINTER(struct_sockaddr_in)).contents
        addr = socket.inet_ntop(family, sa.sin_addr)
        sa = ifa.ifa_netmask.contents
        sa = cast(pointer(sa), POINTER(struct_sockaddr_in)).contents
        netmask = socket.inet_ntop(family, sa.sin_addr)
        yield family, {'address': addr, 'netmask': netmask}


def update_inventory(inventory):
    inventory['fqdn'] = socket.getfqdn()
    inventory['hostname'] = socket.gethostname()

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(('example.com', 1))  # connect() for UDP doesn't send packets
    inet_addr = s.getsockname()[0]

    ifap = POINTER(struct_ifaddrs)()
    result = libc.getifaddrs(pointer(ifap))
    if result != 0:
        raise RuntimeError(get_errno())
    families = {2: 'AF_INET', 10: 'AF_INET6'}
    ifaces = inventory['ifaces'] = {}
    try:
        for ifa in iter_ifaps(ifap):
            name = ifa.ifa_name.decode('utf8')
            if name == 'lo':
                continue
            d = ifaces.setdefault(name, {
                'index': libc.if_nametoindex(ifa.ifa_name),
                'primary': False,
                'name': name,
            })
            for family, addr in getinfos(ifa):
                if addr:
                    family = families[family][3:].lower()
                    if family == 'inet' and addr['address'] == inet_addr:
                        d['primary'] = True
                    d.setdefault(family, []).append(addr)
                    filename = '/sys/class/net/{0}/address'.format(name)
                    if os.path.isfile(filename):
                        with codecs.open(filename, 'r', 'utf8') as fd:
                            d['macaddress'] = fd.read().strip() or None
        return ifaces
    finally:
        libc.freeifaddrs(ifap)


def finalize_inventory(inventory):
    import ipaddress
    ifaces = inventory['ifaces']
    for iface in ifaces.values():
        for net in iface.get('inet', []):
            try:
                addr = ipaddress.IPv4Address(net['address'])
            except ValueError:
                is_private = None
            else:
                is_private = addr.is_private
            net['is_private'] = is_private


if __name__ == '__main__':
    inventory = {}
    update_inventory(inventory)
    finalize_inventory(inventory)
    print(inventory)



# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
import time
import asyncio
import resource
from operator import itemgetter
from collections import deque
from collections import OrderedDict

import nuka
from nuka import log
from nuka import process
from nuka.task import wait_for_boot
from nuka.task import get_task_from_stack
from nuka.task import destroy as destroy_task

RLIMIT_NOFILE = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
resource.setrlimit(resource.RLIMIT_NOFILE, (RLIMIT_NOFILE, RLIMIT_NOFILE))
MAX_PROCESSES = int(RLIMIT_NOFILE / 4)


class HostGroup(OrderedDict):
    """A dict like object to group hosts"""

    async def boot(self):
        raise NotImplementedError()

    async def destroy(self):  # pragma: no cover
        hosts = list(self.values())
        if hosts:
            return await asyncio.wait([destroy_task(h) for h in hosts])

    def __repr__(self):
        return repr([k for k in self])


all_hosts = HostGroup()
nuka.config['all_hosts'] = all_hosts


class TimeIt(object):

    def __init__(self, host, task=None, **kwargs):
        if task is None:  # pragma: no cover
            task = get_task_from_stack()
        kwargs['task'] = task
        self.start = None
        self.host = host
        self.kwargs = kwargs

    def __enter__(self, *args, **kwargs):
        self.start = time.time()

    def __exit__(self, *args, **kwargs):
        self.host.add_time(start=self.start, **self.kwargs)


class BaseHost(object):

    provider = None
    processes_count = 0
    stds = dict(
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    def __init__(self, hostname=None, port='22', **vars):
        if 'address' in vars:
            self.name = self.hostname = vars.pop('address')
        else:
            self.name = hostname.split('.', 1)[0]
            self.hostname = hostname
        self.name = vars.get('name', self.name)
        self.vars = vars
        self.max_sessions = int(self.vars.pop('max_sessions', 10))
        self.vars.setdefault('user', 'root')
        self.vars.setdefault('port', '22')
        self.vars.setdefault('use_sudo', False)
        self.vars.setdefault('archive_modes', ('x:gz',))

        self.loop = vars.pop('loop', asyncio.get_event_loop())

        self._sessions = deque()
        self._cancelled = False
        self._failed = None
        self._start = time.time()
        self._processes = {}
        self._tasks = []
        self._named_tasks = {}
        self._task_times = []
        self._start = time.time()
        self._log = None

        logger = self.vars.pop('logger', None)
        if logger is not None:  # pragma: no cover
            self._log = logger

        self.fully_booted = asyncio.Future(loop=self.loop)

        all_hosts[self.name] = self

    @property
    def log(self):
        if self._log is None:
            self._log = log.HostLogger(self)
        return self._log

    def add_task(self, task):
        self._tasks.append(task)

    def running_tasks(self):
        return [t for t in self._tasks if t.running()]

    def timeit(self, task=None, **kwargs):
        return TimeIt(self, task=task, **kwargs)

    def add_time(self, start=None, task=None, **kwargs):
        kwargs.setdefault('time', time.time() - start)
        if task is None:  # pragma: no cover / maybe no longer required
            task = get_task_from_stack()
        if task is not None:
            kwargs.update(start=start, task=task)
            self._task_times.append(kwargs)
        else:  # pragma: no cover
            self.log.warning("can't retrieve task\n{}".format(kwargs))

    def cancel(self):
        for task in self.running_tasks():
            if not task.done():  # pragma: no cover
                task.cancel()
        self._cancelled = True

    def cancelled(self):
        return self._cancelled

    def fail(self, exc):
        if not self._failed:
            self.cancel()
            self._failed = exc

    def failed(self):
        return self._failed

    def _get_best_addresses(self, public=True):
        hvars = self.vars
        key = public and 'public_ip' or 'private_ip'
        try:
            return hvars[key]
        except KeyError:
            pass
        ifaces = hvars.get('inventory', {}).get('ifaces', {})
        for iface in sorted(ifaces.values(), key=itemgetter('index')):
            if not iface.get('macaddress'):
                # tunX
                continue
            for net in iface.get('inet', []):
                if iface['primary']:
                    if net['is_private']:
                        hvars['private_ip'] = net['address']
                    else:
                        hvars['public_ip'] = net['address']
                elif not net['is_private'] and 'public_ip' not in hvars:
                    hvars['public_ip'] = net['address']
                elif net['is_private'] and 'private_ip' not in hvars:
                    hvars['private_ip'] = net['address']
        return hvars.get(key)

    @property
    def public_ip(self):
        """return host's public ip"""
        return self._get_best_addresses(public=True)

    @property
    def private_ip(self):
        """return host's private ip"""
        return self._get_best_addresses(public=False)

    def __getattr__(self, attr):
        return self.vars[attr]

    def __str__(self):
        return self.name

    def __repr__(self):
        s = '<{0} {1}'.format(self.__class__.__name__, self.name)
        if self.cancelled():
            s += ' cancelled'
        s += '>'
        return s

    @property
    def bootstrap_command(self):
        return self.vars.get('bootstrap_command')

    async def boot(self):  # pragma: no cover
        """boot the host"""
        return dict(rc=0)

    async def get_inventory(self):  # pragma: no cover
        """return host's inventory. await for host's boot & setup if needed"""
        if not self.fully_booted.done():
            await wait_for_boot(self)
        return self.vars['inventory']

    async def destroy(self):  # pragma: no cover
        """destroy the host"""
        self.vars['destroyed'] = True
        return dict(rc=0)

    async def acquire_session_slot(self):
        while self.processes_count > MAX_PROCESSES:
            self.log.debug5('wait for free fds')
            await asyncio.sleep(.5, loop=self.loop)
        sessions = self._sessions
        ll = len(sessions)
        if ll >= self.max_sessions:  # pragma: no cover
            self.log.debug5('wait for a session')
            while ll >= self.max_sessions:
                if not self.cancelled():
                    await asyncio.sleep(.5, loop=self.loop)
                else:
                    return sessions
                ll = len(sessions)
        self.__class__.processes_count += 1
        sessions.append(1)

    def free_session_slot(self):
        self.__class__.processes_count -= 1
        self._sessions.pop()

    async def create_process(self, cmd, task=None, **kwargs):
        if self.cancelled():
            raise asyncio.CancelledError()
        process_cmd = self.wraps_command_line(cmd, **kwargs)
        proc = await process.create(process_cmd, self, task)
        return proc

    async def run_command(self, cmd=None, stdin=None, task=None, **kwargs):
        """run a shell command on the remote host"""
        proc = await self.create_process(cmd, task)
        if stdin:
            proc.stdin.write(stdin)
            await proc.stdin.drain()
            # close_stdin is not recommended. we cant send signals after that
            # but it's usefull for testing
            if kwargs.get('close_stdin'):
                proc.stdin.close()
        stdout, stderr = await asyncio.gather(proc.stdout.read(),
                                              proc.stderr.read(),
                                              loop=self.loop)
        if kwargs.get('wait', True):
            await proc.wait()
        return dict(rc=proc.returncode, stdout=stdout, stderr=stderr)

    async def send_messages(self, message):  # pragma: no cover
        coros = []
        for proc in self._processes.values():
            try:
                coro = proc.send_message(message)
            except ConnectionResetError:
                pass
            else:
                coros.append(coro)
        if coros:
            try:
                await asyncio.wait(coros)
            except ConnectionResetError:
                pass

    @classmethod
    def from_stdin(cls):
        for line in sys.stdin:
            yield cls(hostname=line.strip())


class Host(BaseHost):
    """A host. Used by tasks as target"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not os.getenv('SSH_AUTH_SOCK'):
            self.log.warning('No SSH_AUTH_SOCK set. Your tasks may freeze')

    def wraps_command_line(self, cmd, **kwargs):
        ssh_user = kwargs.get('switch_ssh_user')
        if ssh_user is None:
            # we use the main user account
            switch_user = kwargs.get('switch_user') or 'root'
            if switch_user != 'root':
                if switch_user != self.vars['user']:
                    # we have to use sudo
                    args = (switch_user, cmd)
                    if self.use_sudo:
                        cmd = '{sudo} -u {0} {1}'.format(*args, **nuka.config)
                    else:
                        cmd = '{su} -c "{1}" {0}'.format(*args, **nuka.config)
            elif self.use_sudo:
                cmd = '{sudo} {0}'.format(cmd, **nuka.config)

        if ssh_user is None:
            ssh_user = self.vars['user']

        ssh_cmd = ['ssh'] + nuka.config['ssh']['options'] + ['-l', ssh_user]
        if self.port:
            ssh_cmd.extend(['-p', self.port])
        ssh_cmd.extend([self.hostname, cmd])
        return ssh_cmd


class LocalHost(BaseHost):

    def __init__(self):
        super().__init__(hostname='localhost')

    def wraps_command_line(self, cmd, **kwargs):
        ssh_cmd = ['bash', '-c', cmd]
        return ssh_cmd


class Chroot(BaseHost):

    def __init__(self, path):
        super().__init__(hostname=path.split('/')[-1])
        self.path = path

    def wraps_command_line(self, cmd, **kwargs):
        ssh_cmd = ['chroot', self.path, 'bash', '-c', cmd]
        return ssh_cmd



# -*- coding: utf-8 -*-
import jinja2
import pytest
import asyncio
from nuka.tasks import file


@pytest.mark.asyncio
async def test_cat_doc(host):
    res = await file.cat('/etc/default/useradd')
    assert res.content


@pytest.mark.asyncio
async def test_exists_doc(host):
    res = await file.exists('/tmp')
    assert bool(res) is True

    res = await file.exists('/nope')
    assert bool(res) is False


@pytest.mark.asyncio
async def test_mkdir_doc(host):
    if not await file.exists('/tmp/doc'):
        await file.mkdir('/tmp/doc')


@pytest.mark.asyncio
async def test_rm_doc(host):
    await file.rm('/tmp/doc')


@pytest.mark.asyncio
async def test_put_doc(host):
    await file.put([
        dict(src='/etc/resolv.conf', dst='/tmp/resolv.conf'),
        dict(src='docs/utils.py', dst='/tmp/utils.py', executable=True),
        # jinja2 template
        dict(src='example.j2', dst='/tmp/xx1', mod='600'),
        # symlink
        dict(linkto='/etc/hosts', dst='/etc/hosts2'),
    ], ctx=dict(name='example'))


@pytest.mark.asyncio
async def test_update_doc(host):
    await file.update(
        dst='/etc/default/useradd',
        replaces=[(r'^\# HOME=/home', 'HOME=/new_home')])


@pytest.mark.asyncio
async def test_mkdir_rmdir(host):
    res = await file.exists('/tmp/tox_test')
    assert bool(res) is False
    res = await file.mkdir(dst='/tmp/tox_test')
    assert bool(res)
    res = await file.exists(dst='/tmp/tox_test')
    assert bool(res)
    res = await file.rm(dst='/tmp/tox_test')
    assert bool(res)
    res = await file.exists(dst='/tmp/tox_test')
    assert bool(res) is False


@pytest.mark.asyncio
async def test_mkdir_rmdir_unauthorized(host, user):
    with pytest.raises(asyncio.CancelledError):
        await file.mkdir(dst='/etc/tox_test', switch_user=user)
    with pytest.raises(asyncio.CancelledError):
        await file.mkdir(dst='/var', switch_user=user)


@pytest.mark.asyncio
async def test_put(host):
    with open('/tmp/to_put.txt', 'wb') as fd:
        fd.write(b'yo')
    res = await file.put([dict(
        src='/tmp/to_put.txt', dst='/tmp/xx', executable=True
    )])
    assert bool(res)
    res = await file.cat('/tmp/xx')
    assert res.content == 'yo'


@pytest.mark.asyncio
async def test_put_with_user(host, user):
    with open('/tmp/to_put.txt', 'wb') as fd:
        fd.write(b'yo')
    res = await file.put([dict(
        src='/tmp/to_put.txt', dst='/tmp/test_' + user,
        )],
        switch_user=user)
    assert bool(res)
    res = await file.rm(dst='/tmp/test_' + user,
                        switch_user=user)
    assert bool(res)

    with pytest.raises(asyncio.CancelledError):
        await file.put([dict(src='/tmp/to_put.txt', dst='/etc/test_' + user)],
                       switch_user=user)


@pytest.mark.asyncio
async def test_scripts(host):
    hid = str(id(host))
    script = '#!/bin/bash\necho {0} > /tmp/h'.format(hid)
    with open('/tmp/to_put_script', 'wb') as fd:
        fd.write(script.encode('utf8'))
    res = await file.scripts([dict(
        src='/tmp/to_put_script', dst='/tmp/script'
    )])
    assert bool(res)

    res = await file.cat('/tmp/h')
    assert res.content.strip() == hid


@pytest.mark.asyncio
async def test_template(host):
    res = await file.put([
        dict(src='example.j2', dst='/tmp/xx0', executable=True),
        dict(src='example.j2', dst='/tmp/xx1', mod='600'),
        ],
        ctx=dict(name='dude'))
    assert bool(res)
    for i in range(0, 2):
        res = await file.cat('/tmp/xx{0}'.format(i))
        assert res.content == 'yo dude\n'


@pytest.mark.asyncio
async def test_template_error(host):
    with pytest.raises(jinja2.UndefinedError):
        await file.put([dict(src='error.j2', dst='/tmp/xx0')])


@pytest.mark.asyncio
async def test_update(host):
    res = await file.update(
        dst='/etc/default/useradd',
        replaces=[(r'^.*HOME=/home', 'HOME=/new_home')])
    assert bool(res)

    res = await file.cat('/etc/default/useradd')
    assert 'HOME=/new_home' in res.content

    res = await file.update(
        dst='/etc/default/useradd',
        replaces=[(r'HOME=/new_home', 'HOME=/home')])
    assert bool(res)

    res = await file.cat('/etc/default/useradd')
    assert 'HOME=/new_home' not in res.content


@pytest.mark.asyncio
async def test_mv(host):
    with open('/tmp/to_move.txt', 'wb') as fd:
        fd.write(b'yo')
    res = await file.put([dict(
        src='/tmp/to_move.txt', dst='/tmp/to_move.txt',
        )])
    res = await file.mv(src='/tmp/to_move.txt', dst='/tmp/moved.txt')
    assert bool(res)
    res = await file.cat('/tmp/moved.txt')
    assert res.content.strip() == 'yo'



# -*- coding: utf-8 -*-
from nuka.hosts import DockerContainer
from nuka.tasks import shell
import nuka

host = DockerContainer(hostname='debian', image='bearstech/nukai')


async def my_tasks(host):
    await shell.shell('whoami')

nuka.run(my_tasks(host))



# -*- coding: utf-8 -*-
"""
unix user related tasks
"""
import os
from nuka import utils
from nuka.task import Task
from nuka.tasks import file


class delete_user(Task):
    """remove a unix account"""

    def __init__(self, username=None, **kwargs):
        kwargs.setdefault('name', username)
        super(delete_user, self).__init__(username=username, **kwargs)

    def do(self):
        username = self.args['username']
        if not self.sh(['id', username], check=False)['rc']:
            return self.sh(['userdel', '-r', username])
        return dict(rc=0, username=username)

    def diff(self):
        username = self.args['username']
        diff = ''
        if not self.sh(['id', username], check=False)['rc']:
            diff = self.texts_diff(username, '')
        return dict(rc=0, diff=diff, username=username)


class create_user(Task):
    """create a unix account"""

    home_format = '/home/{0}'
    gecos_format = 'user {0}'

    def __init__(self, username=None, **kwargs):
        kwargs.setdefault('name', username)
        kwargs.setdefault('gecos', self.gecos_format.format(username))
        kwargs.setdefault('home', self.home_format.format(username))
        super(create_user, self).__init__(username=username, **kwargs)

    def do(self):
        username = self.args['username']
        home = self.args['home']
        gecos = self.args['gecos']
        if self.sh(['id', username], check=False)['rc']:
            if self.is_debian:
                # ensure adduser is installed (wheezy do not have it)
                res = self.sh(['which', 'adduser'], check=False)
                if res['rc'] != 0:  # pragma: no cover
                    raise OSError('adduser is not available')
                cmd = [
                    'adduser', '--quiet', '--disabled-password',
                    '--gecos', gecos, '--home', home, username]
            else:
                cmd = [
                    'useradd',  '-p', '\*', '-c', gecos,
                    '-m',  '--home-dir', home, username]
            res = self.sh(cmd)
        else:
            res = dict(rc=0, changed=False)
        res.update(username=username, home=home)
        return res

    def diff(self):
        username = self.args['username']
        home = self.args['home']
        diff = ''
        if self.sh(['id', username], check=False)['rc'] != 0:
            diff = self.texts_diff('', username)
        if not os.path.isdir(home):
            diff += self.texts_diff('', home)
        res = dict(rc=0, diff=diff, **self.args)
        res.update(username=username, home=self.args['home'])
        return res


class create_www_user(create_user):
    """create a unix account with /var/www/{username} as home"""

    home_format = '/var/www/{0}'
    gecos_format = 'www user {0}'


class authorized_keys(file.put):
    """upload ssh keys from string or file"""

    def __init__(self, username=None, keys=None, keysfile=None, **kwargs):
        kwargs.update(name=username, username=username, switch_user=username)
        if 'files' not in kwargs:
            dst = '~/.ssh/authorized_keys'
            if keysfile:
                kwargs['files'] = [dict(src=keysfile, dst=dst, mod='644')]
            elif keys:
                kwargs['files'] = [dict(data=keys, dst=dst, mod='644')]
            else:
                raise RuntimeError('You must provide keys or keysfile')
        super(authorized_keys, self).__init__(**kwargs)

    def do(self):
        utils.makedirs(os.path.expanduser('~/.ssh'), mod='700')
        super(authorized_keys, self).do()



# -*- coding: utf-8 -*-
import os
import pytest
import tempfile

from nuka.tasks import file
from nuka.tasks import virtualenv as venv

pytestmark = [
    pytest.mark.skipif(
        'python2' in os.environ['ENV_NAME'], reason='need a recent pip'),
]


@pytest.mark.asyncio
async def test_virtualenv_doc(host):
    res = await venv.virtualenv('/tmp/venv')
    assert res


@pytest.mark.asyncio
async def test_requirements(host, diff_mode):
    if await file.exists('/tmp/venv'):
        await file.rm('/tmp/venv')

    with diff_mode:
        res = await venv.virtualenv('/tmp/venv')
        assert '+/tmp/venv/bin/python' in res.res['diff'], res.res['diff']

    with tempfile.NamedTemporaryFile() as fd:
        fd.write(b'six')
        fd.flush()
        assert await venv.virtualenv('/tmp/venv', requirements=fd.name)

    with diff_mode:
        res = await venv.virtualenv('/tmp/venv')
        assert '+/tmp/venv/bin/python' not in res.res['diff'], res.res['diff']



'''
Usage: cat hostlist | python from_stdin.py
'''
import nuka
from nuka.tasks import file
from nuka.hosts import Host


async def do(host):
    res = await file.cat('/etc/debian_version')
    print(res.content)

nuka.run_all(do, *Host.from_stdin())



# -*- coding: utf-8 -*-
from nuka.task import Task


class psql(Task):
    """psql"""

    diff = False

    def __init__(self, sql=None, **kwargs):
        kwargs.setdefault('switch_user', 'postgres')
        super(psql, self).__init__(sql=sql, **kwargs)

    def do(self):
        return self.sh(['psql', '-tAc', self.args['sql']])


class create_db(Task):
    """create a database and grant user"""

    def __init__(self, name=None, user=None, password=None, **kwargs):
        kwargs.setdefault('switch_user', 'postgres')
        kwargs.setdefault('template', 'template1')
        kwargs.setdefault('locale', 'en_US.UTF-8')
        super(create_db, self).__init__(name=name, user=user,
                                        password=password, **kwargs)

    def user_exists(self):
        s = "SELECT 1 FROM pg_roles WHERE rolname='{user}'".format(**self.args)
        res = self.sh(['psql', '-tAc', s], check=False)
        return res['stdout'].strip() == "1"

    def db_exists(self):
        res = self.sh(['psql', self.args['name'], '-c', ''], check=False)
        return res['rc'] == 0

    def do(self):
        if not self.db_exists():
            if not self.user_exists():
                s = "CREATE ROLE {user} WITH PASSWORD '{password}';"
                self.sh(['psql', '-tAc', s.format(**self.args)])
            return self.sh([
                'createdb',
                '--owner', self.args['user'],
                '--template', self.args['template'],
                '--lc-ctype', self.args['locale'],
                '--lc-collate', self.args['locale'],
                self.args['name'],
            ])
        else:
            return dict(rc=0, changed=False)

    def diff(self):
        diff = []
        if not self.user_exists():
            diff.append(self.texts_diff('', self.args['user'],
                                        tofile='role'))
        if not self.db_exists():
            diff.append(self.texts_diff('', self.args['name'],
                                        tofile='database'))
        if diff:
            return dict(diff='\n'.join(diff))
        return dict(rc=0, changed=False)



# -*- coding: utf-8 -*-
import pytest

from nuka.task import Task


@pytest.mark.asyncio
async def test_error(host):

    class ErrTask(Task):
        def pre_process(self):
            raise OSError()

    with pytest.raises(OSError):
        await ErrTask()



# -*- coding: utf-8 -*-
"""
python virtualenv
"""
import os
import sys
from nuka import utils
from nuka.tasks import file
from nuka.tasks import http

GET_PIP = 'https://bootstrap.pypa.io/get-pip.py'


def d():
    import logging
    with open('/tmp/nuka_provisionning/nuka/bin/coverage') as fd:
        logging.warn(fd.read())


class virtualenv(file.put):
    """create a python virtualenv"""

    def __init__(self, dst=None, requirements=None, upgrade=False, **kwargs):
        kwargs.setdefault('name', dst)
        kwargs.setdefault('requirements', requirements)
        super(virtualenv, self).__init__(dst=dst, **kwargs)

    def pre_process(self):
        if self.args['requirements']:
            self.args['files'] = [dict(
                src=self.args['requirements'],
                dst=os.path.join(self.args['dst'],
                                 os.path.basename(self.args['requirements'])))]
            super(virtualenv, self).pre_process()

    def do(self):
        dst = self.args['dst']
        pip = os.path.join(dst, 'bin', 'pip')
        binary = os.path.join(dst, 'bin', 'python')
        if not os.path.isfile(binary):
            executable = self.args.get('executable',
                                       utils.best_executable())
            if 'python3' in executable:
                # use stdlib venv
                args = [dst]
                self.sh([executable, '-m', 'venv', '--without-pip'] + args)
                self.sh([binary, '-m', 'ensurepip'], check=False)
            else:
                # use virtualenv
                dirname = os.path.dirname(sys.executable)
                venv = os.path.join(dirname, 'virtualenv')
                if not os.path.isfile(venv):
                    venv = 'virtualenv'

                res = self.sh([venv, '-p', executable, dst])

        if not os.path.isfile(binary):
            raise OSError('Not able to create {0}'.format(binary))

        has_pip = False
        if os.path.isfile(pip):
            has_pip = True
        if not has_pip:
            res = self.sh([binary, '-m', 'pip', '--version'], check=False)
            if res['rc'] == 0:
                has_pip = True

        if not has_pip:
            res = self.check(http.fetch(
                src=GET_PIP,
                dst=os.path.join(dst, 'get-pip.py')).do())
            res = self.sh([binary, res['dst']])
        else:
            res = dict(rc=0)

        files = self.args.get('files') or []
        if files:
            self.check(super(virtualenv, self).do())
            for f in files:
                if os.path.isfile(pip):
                    cmd = [pip, 'install']
                else:
                    cmd = [binary, '-m', 'pip', 'install']
                if self.args.get('upgrade'):
                    cmd.append('--upgrade')
                cmd.extend(['-r', f['dst']])
                self.sh(cmd)
        res.update(python=binary)
        return res

    def diff(self):
        dst = self.args['dst']
        binary = os.path.join(dst, 'bin', 'python')
        diff = ''
        if not os.path.isfile(binary):
            diff += self.texts_diff('', binary, fromfile=dst)
        res = super(virtualenv, self).diff()
        diff += res.get('diff', '')
        return dict(rc=0, diff=diff, python=binary)



# -*- coding: utf-8 -*-
from functools import partial
import subprocess
import asyncio
import sys

import docker as docker_py

from nuka.task import get_task_from_stack
from nuka.hosts.base import BaseHost
from nuka.hosts.base import HostGroup
from nuka.task import wait_for_boot
import nuka

try:
    DockerClient = docker_py.Client
except AttributeError:
    DockerClient = docker_py.APIClient


class DockerContainer(BaseHost):
    """By default the image will be ``bearstech/nukai:latest`` which is the
    latest debian with python3 installed but you can use whatever you want. We
    also provide a bunch of `nukai(mages)
    <https://hub.docker.com/r/bearstech/nukai/tags/>`_

    .. code-block:: py

        >>> host = DockerContainer(
        ...     hostname='myhost',
        ...     image='bearstech/nukai:debian-jessie-python3')
    """

    provider = 'docker'

    def __init__(self, hostname=None, image='bearstech/nukai:latest',
                 **kwargs):
        kwargs.update(hostname=hostname, image=image)
        super().__init__(**kwargs)
        self.cli = DockerClient()

    @property
    def bootstrap_command(self):
        if 'bootstrap_command' in self.vars:
            return self.vars['bootstrap_command']
        elif self.vars['image'].startswith('debian'):
            return (
                'which python || (apt-get update -qq > /dev/null; '
                ' apt-get install -qqy '
                'python-virtualenv wget perl-modules adduser)'
            )
        elif self.vars['image'].startswith('centos'):
            return (
                'ls /usr/bin/wget 2>&1 > /dev/null || '
                'yum install -y -q -q wget 2>&1 > /dev/null;'
                'ls /usr/bin/virtualenv 2>&1 > /dev/null || '
                'yum install -y -q -q python-virtualenv 2>&1 > /dev/null'
            )

    async def acquire_session_slot(self):
        return

    async def acquire_connection_slot(self):
        return

    def free_session_slot(self):
        return

    def wraps_command_line(self, cmd, **kwargs):
        # there is no ssh in docker so we use the regular switch_user
        ssh_user = kwargs.get('switch_ssh_user')
        if ssh_user is not None:
            kwargs['switch_user'] = ssh_user

        switch_user = kwargs.get('switch_user') or 'root'
        if switch_user != 'root':
            if switch_user != self.vars['user']:
                # we have to use sudo
                args = (switch_user, cmd)
                if self.use_sudo:
                    cmd = '{sudo} -u {0} {1}'.format(*args, **nuka.config)
                else:
                    cmd = '{su} -c "{1}" {0}'.format(*args, **nuka.config)
        elif self.use_sudo:
            cmd = '{sudo} {0}'.format(cmd, **nuka.config)

        cmd = ['docker', 'exec', '-i', str(self), 'bash', '-c', cmd]
        return cmd

    async def boot(self):
        # we need to get the task from here.
        # we cant retrieve it while in an executor
        task = get_task_from_stack()
        boot = partial(self._boot_api, task=task)
        return await self.loop.run_in_executor(None, boot)

    def _boot_api(self, task=None):
        with self.timeit(type='api_call', task=task, name='start()'):
            try:
                self.cli.start(self.name)
            except docker_py.errors.APIError:
                pass
        with self.timeit(type='api_call', task=task, name='containers()'):
            containers = self.cli.containers(filters=dict(name=self.name))
        if containers:
            container = containers[0]
        else:
            with self.timeit(type='api_call', task=task, name='images()'):
                found = False
                for image in self.cli.images():
                    if image['RepoTags'] == [self.image]:
                        found = True
            if not found:
                with self.timeit(type='api_call', task=task, name='pull()'):
                    self.log.warning('Pulling image {0}...'.format(self.image))
                    subprocess.call(['docker', 'pull', self.image])
            self.log.debug('Create container...')
            with self.timeit(type='api_call', task=task, name='create()'):
                container = self.cli.create_container(
                    image=self.image,
                    name=self.name, hostname=self.name,
                    command=self.vars.get('command', None))
        self.vars['container'] = container
        self.vars['container_id'] = container['Id']
        try:
            self.cli.start(self.name)
        except docker_py.errors.APIError:
            pass
        else:
            self.log.debug('Container started'.format(self))
        return container

    @property
    def private_ip(self):
        if 'private_ip' not in self.vars:
            container = self.vars['container']
            net = list(container['NetworkSettings']['Networks'].values())[0]
            ip = net['IPAddress']
            # containers has no public ip
            self.vars['public_ip'] = self.vars['private_ip'] = ip
        return self.vars['private_ip']
    public_ip = private_ip

    async def destroy(self):
        remove_container = partial(self.cli.remove_container,
                                   self.hostname, force=True)
        await self.loop.run_in_executor(None, remove_container)


class DockerCompose(HostGroup):
    """A HostGroup that use the docker-compose.yml to provide some hosts:

    .. code-block:: py

        >>> hosts = DockerCompose(project_name='myproject')
    """

    def __init__(self, project_name=None, compose_file=None):
        self.project_name = project_name
        self.compose_file = compose_file
        self.args = []
        if self.project_name:
            self.args.extend(['--project-name', self.project_name])
        if self.compose_file:
            self.args.extend(['--file', self.compose_file])
        super().__init__()

    async def boot(self):
        """launch ``docker-compose up`` and setup all containers"""
        cmd = [sys.executable, '-m', 'compose'] + self.args + ['up', '-d']
        subprocess.check_call(cmd)
        cmd = [sys.executable, '-m', 'compose'] + self.args + ['ps']
        stdout = subprocess.check_output(
            cmd, stderr=subprocess.STDOUT)
        stdout = stdout.decode('utf8')
        for line in stdout.split('\n')[2:]:
            if line.strip():
                name = line.split()[0]
                self[name] = DockerContainer(hostname=name)
        if self:
            await asyncio.wait([wait_for_boot(h) for h in self.values()])

    async def destroy(self):
        """launch ``docker-compose down``"""
        cmd = [sys.executable, '-m', 'compose'] + self.args + ['down']
        p = subprocess.Popen(cmd)
        p.wait()
        for host in self.values():
            host.vars['destroyed'] = True
        return True


def update_image(base_image):
    container = DockerContainer(
        hostname='nuka',
        image=base_image,
        command=['bash', '-c', 'while true; do sleep 1000000000000; done'],
        )
    nuka.run(wait_for_boot(host=container))
    image, tag = base_image.split(':', 1)
    container.cli.commit(
        container.container_id, repository='nuka_' + image, tag=tag)
    return container


def clean_images():
    cli = docker_py.Client()
    for image in cli.images():
        if image['RepoTags'] == ['<none>:<none>']:
            print('Removing {Id}'.format(**image))
            cli.remove_image(image['Id'])



# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from uuid import uuid4
from operator import itemgetter
from collections import defaultdict

import jinja2

import nuka
from nuka.utils import json
from nuka.task import teardown


def round_dict(d):
    for k, v in list(d.items()):
        if isinstance(v, float):
            d[k + '_str'] = str(round(v, 1))
    return d


class Item(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['uid'] = str(uuid4()).replace('-', '')

    def __getattr__(self, attr):
        return self[attr]


def get_report_data(host):

    all_tasks = {}
    subtasks = defaultdict(list)
    for t in host._task_times:
        task = t['task']
        if not task.cancelled():
            if t['type'] == 'task':
                all_tasks[id(t['task'])] = t
            else:
                subtasks[id(t['task'])].append(t)
    if not all_tasks:  # pragma: no cover
        return

    all_tasks = sorted(all_tasks.values(), key=itemgetter('start'))

    stats = {}

    _start = all_tasks[0]['start']

    tasks = []
    for row, task_dict in enumerate(all_tasks):
        row += 1
        task = task_dict['task']
        m = task_dict
        local_time = m['time']
        task_name = task.__class_name__()

        stat = stats.setdefault(
            task_name,
            {'name': task_name, 'calls': 0,
             'time': 0, 'remote_time': 0})
        stat['calls'] += 1
        stat['time'] += local_time
        stat['remote_time'] += m.get('remote_time', 0.)

        name = '{0}({1})'.format(
                task_name,
                task.args.get('name', repr(task.args))
        )
        t = Item(
            type="task",
            name=jinja2.escape(name),
            short_name=task_name.replace('nuka.tasks.', ''),
            start=(m['start'] - _start),
            time=local_time,
            rc=task.res.get('rc', 0),
            row=row,
            filename=m['filename'],
            lineno=m['lineno'],
            funcs=[],
            remote_calls=[],
        )
        tasks.append(round_dict(t))

        subs = sorted(subtasks[id(task)], key=itemgetter('start'))
        sh_start = None
        for item in subs:
            if item['type'] == 'api_call':
                name = item['name']
            elif item['type'] == 'pre_process':
                name = 'pre_process()'
            elif item['type'] == 'post_process':
                name = 'post_process()'
                item['type'] == 'pre_process'
            else:
                cmd = ' '.join(item['cmd'])
                if nuka.config['script'] in cmd:
                    cmd = 'nuka/script.py'
                name = 'subprocess({0})'.format(cmd)
            func = Item(
                name=name,
                type=item['type'],
                start=(item['start'] - _start),
                time=item['time'],
                parent=t.uid,
                latency=item.get('latency'),
                remote_calls=[],
            )
            if 'meta' in item:
                sh_start = func['start'] + item['latency']
                shs = item['meta']['remote_calls']
                for sh in shs:
                    tsh = Item(
                        type="sh",
                        start=sh_start,
                        name='sh({0})'.format(sh['cmd']),
                        time=sh['time'],
                        parent=t.uid,
                        rc=sh['rc'],
                    )
                    func['remote_calls'].append(round_dict(tsh))
                    sh_start += tsh.time
            t['funcs'].append(round_dict(func))

    _end = all_tasks[-1]
    _real_end = _end
    i = 1
    while isinstance(_real_end['task'], teardown):
        i += 1
        try:
            _real_end = all_tasks[-i]
        except IndexError:
            _real_end = _end
            break

    for k, v in stats.items():
        v['avg_time'] = v['time'] / v['calls']
        v['avg_remote_time'] = v['remote_time'] / v['calls']
        for kk, vv in v.items():
            if isinstance(vv, float):
                v[kk] = '%.3f' % vv
    stats = sorted(stats.values(),
                   key=itemgetter('remote_time'),
                   reverse=True)

    data = {
        'host': host.name,
        'tasks': tasks,
        'stats': stats,
        'total_time': _end['start'] + _end['time'] - _start,
        'real_time': _real_end['start'] + _real_end['time'] - _start,
    }
    return data


def build_reports(hosts):
    hosts_data = {}
    for host in hosts:
        data = get_report_data(host)
        if data:
            hosts_data[str(host)] = round_dict(data)

    if not hosts_data:
        return

    dirname = nuka.config['reports']['dirname']
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    data = {'hosts': hosts_data, 'total_time': 0}
    data['total_time'] = max([d['total_time'] for d in hosts_data.values()])
    data['total_tasks'] = max([len(d['tasks']) for d in hosts_data.values()])

    engine = nuka.config.get_template_engine()

    report_name = nuka.config['reports'].get('name')
    if report_name is None:
        filename = os.path.split(sys.argv[0])[-1]
        report_name = os.path.splitext(filename)[0]

    filename = os.path.join(dirname, '{0}.json'.format(report_name))
    dumped_data = json.dumps(data, indent=2)
    with open(filename, 'w') as fd:
        fd.write(dumped_data)

    ctx = dict(
        data=data,
        dumped_data=dumped_data,
    )

    filename = os.path.join(dirname, '{0}_gantt.html'.format(report_name))
    template = engine.get_template('reports/gantt.html.j2')
    with open(filename, 'w') as fd:
        fd.write(template.render(ctx))

    filename = os.path.join(dirname, '{0}_stats.html'.format(report_name))
    template = engine.get_template('reports/stats.html.j2')
    with open(filename, 'w') as fd:
        fd.write(template.render(ctx))


def run_dev_server():  # pragma: no cover
    """Usage: python -m nuka.report <type> <data_filename>"""
    from wsgiref.simple_server import make_server
    engine = nuka.config.get_template_engine()
    type, filename = sys.argv[-2:]

    def app(environ, start_response):
        start_response('200 OK',
                       [('Content-Type', 'text/html; charset=utf8')])
        template = engine.get_template('reports/{0}.html.j2'.format(type))
        with open(filename) as fd:
            data = json.load(fd)
        html = template.render(dict(data=data, dumped_data=json.dumps(data)))
        return [html.encode('utf8')]

    httpd = make_server('', 8000, app)
    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':  # pragma: no cover
    run_dev_server()



# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.

import os
import yaml
import subprocess

import jinja2
from jinja2.utils import open_if_exists
from jinja2.loaders import split_template_path
from jinja2.exceptions import TemplateNotFound


class FileSystemLoader(jinja2.FileSystemLoader):

    def get_source(self, environment, template):
        pieces = split_template_path(template)
        for searchpath in self.searchpath:
            filename = os.path.join(searchpath, *pieces)
            f = open_if_exists(filename)
            if f is None:
                continue
            if filename.endswith('.gpg'):
                f.close()
                _, contents = decrypt(filename, self.encoding)
            else:
                try:
                    contents = f.read().decode(self.encoding)
                finally:
                    f.close()

            mtime = os.path.getmtime(filename)

            def uptodate():
                try:
                    return os.path.getmtime(filename) == mtime
                except OSError:
                    return False
            return contents, filename, uptodate
        raise TemplateNotFound(template)


def decrypt(filename, encoding='utf8'):
    cmd = ['gpg', '--quiet', '--batch', '-d', filename]
    try:
        value = subprocess.check_output(cmd)
    except subprocess.CalledProcessError:
        raise
        raise RuntimeError((
            'Error while trying to decrypt {0}. '
            'Maybe your GPG agent has expired'.format(filename)))
    if filename.endswith('.gpg'):
        filename = filename[:-4]
    if filename.endswith(('.yaml', '.yml')):
        return yaml.load(value)
    if isinstance(value, bytes):
        value = value.decode(encoding)
    return filename, value



# -*- coding: utf-8 -*-
import nuka
from nuka.hosts import DockerContainer

from tasks.timezone import timezone

host = DockerContainer(hostname='debian_jessie')


async def change_timezone(host):
    await timezone(tz='Europe/Paris')


nuka.run(change_timezone(host))



# -*- coding: utf-8 -*-
import nuka
from nuka.hosts import DockerContainer
from nuka.tasks import shell

host = DockerContainer(
    hostname='debian_jessie',
    image='debian:jessie',
    command=['bash', '-c', 'while true; do sleep 1000000000000; done'],
)


async def sleep(host):
    await shell.command(['sleep', '5'], watch=1)


nuka.run(sleep(host))



# -*- coding: utf-8 -*-
import nuka
from nuka.hosts import DockerContainer
from nuka.tasks import shell

host = DockerContainer(hostname='nukai')

e = nuka.Event('event')


@nuka.cancel_on_error(e)
async def fail(host):
    await shell.command(['sleep', '2'])
    raise AttributeError()
    await shell.command(['sleep', '1'])
    e.release()


async def wait(host):
    await nuka.wait(e)
    await shell.command(['sleep', '2'])


async def not_wait(host):
    await shell.command(['sleep', '1'])
    await shell.command(['rm', '/usr'])
    await shell.command(['sleep', '3'])

nuka.run(
    fail(host),
    wait(host),
    not_wait(host),
)
nuka.run(host.destroy())



# -*- coding: utf-8 -*-
from collections import defaultdict
from functools import partial
import ipaddress
import threading
import asyncio
import codecs
import time
import os

from libcloud.common.google import ResourceNotFoundError
from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from novaclient.client import Client as NovaClient
from novaclient.exceptions import NotFound

from nuka.task import get_task_from_stack
from nuka.hosts import base
from nuka import utils
import nuka


nova_providers = (Provider.OPENSTACK,)


_drivers = defaultdict(dict)

_env_keys = {
    Provider.GCE: (
        ('GCE_EMAIL', 'user_id'),
        ('GCE_PEM_FILE_PATH', 'key'),
        ('GCE_PROJECT', 'project'),
    ),
    Provider.OPENSTACK: (
        ('OS_VERSION', '2.0'),
        ('OS_AUTH_URL', 'auth_url'),
        ('OS_USERNAME', 'username'),
        ('OS_PASSWORD', 'password'),
        ('OS_PROJECT_NAME', 'project_name'),
        ('OS_PROJECT_ID', 'project_id'),
        ('OS_TENANT_NAME', 'project_name'),
        ('OS_TENANT_ID', 'project_id'),
        ('OS_REGION_NAME', 'region_name'),
    )
}


def driver_from_config(provider, **args):
    # cache one driver per thread. probably only usefull for main thread
    ident = threading.get_ident()
    driver = _drivers[provider].get(ident)
    if driver is None:
        if provider in nova_providers:
            klass = NovaClient
        else:
            klass = get_driver(provider)

        if not args:
            args = nuka.config[provider.lower()].get('driver', {}) or {}
            if isinstance(args, str):
                # load arguments from file
                with open(os.path.expanduser(args)) as fd:
                    args = utils.json.load(fd)

            for k, new_k in _env_keys.get(provider, ()):
                if k in os.environ:
                    v = os.environ[k]
                    args[new_k] = v
            if not args:
                raise RuntimeError((
                    'Not able to get the driver configuration for {} provider.'
                ).format(provider))

        driver = klass(**args)
        _drivers[ident] = driver
    return driver


class Host(base.Host):
    """Host in the cloud"""

    use_sudo = True

    def __init__(self, hostname, node=None,
                 create=True, driver_args=None, create_node_args=None, **vars):
        config = nuka.config.get(self.provider.lower(), {})
        user = config.get('user') or 'root'
        vars.setdefault('user', user)
        vars['driver_args'] = driver_args or {}
        vars['create_node_args'] = create_node_args or {}
        super().__init__(hostname, **vars)
        self.create = create
        self._node = node

    async def boot(self):
        # we need to get the task from here.
        # we cant retrieve it while in an executor
        task = get_task_from_stack()
        get_node = partial(self.get_node, task=task)
        try:
            self._node = await self.loop.run_in_executor(None, get_node)
        except asyncio.CancelledError:
            raise
        except Exception as e:
            print(type(e))
            self.log.exception('boot')
            self.cancel()
            raise
        else:
            self.hostname = self.public_ip

    @property
    def public_ip(self):
        """return host's public ip"""
        return self.node.public_ips[0]

    @property
    def private_ip(self):
        """return host's private ip"""
        return self.node.private_ips[0]

    def get_node(self, task=None, **kwargs):
        if self._node is False:
            raise RuntimeError('Node {0} was destroyed'.format(self))
        elif self._node is None:

            with self.timeit(type='api_call', task=task,
                             name='Driver.from_config()'):
                driver = driver_from_config(provider=self.provider,
                                            **self.driver_args)

            with self.timeit(type='api_call', task=task,
                             name='driver.ex_get_node()'):
                try:
                    if self.provider in nova_providers:
                        self._node = driver.servers.find(name=self.hostname)
                    else:
                        self._node = driver.ex_get_node(self.hostname)
                except (ResourceNotFoundError, NotFound):
                    pass
            if self._node is None and self.create:
                self.create_node(driver=driver, task=task)
        return self._node
    node = property(get_node)

    def create_node(self, driver=None, task=None, **kwargs):
        self.log.warning(
            'Node {0} does not exist. Creating...'.format(self))
        if driver is None:
            driver = driver_from_config(provider=self.provider,
                                        **self.driver_args)
        args = self.get_create_node_args(driver=driver, task=task)
        args['name'] = self.name
        with self.timeit(type='api_call', task=task,
                         name='driver.ex_create_node()'):
            if self.provider in nova_providers:
                node = driver.servers.create(**args)
                # it take at least 20s to boot a vm
                time.sleep(20)
                wait = 15
                while node.status not in ('ACTIVE',):
                    try:
                        node = driver.servers.find(name=self.hostname)
                    except NotFound:
                        pass
                    if node.status in ('ACTIVE',):
                        # need to wait ~10s to get the server up & running
                        time.sleep(10)
                        break
                    else:
                        wait = (wait - 5)
                        if wait < 1:
                            wait = 2
                        time.sleep(wait)
                self._node = node
            else:
                self._node = driver.create_node(**args)
            self.log.warning('Node {0} created'.format(self))

    def get_create_node_args(self, driver=None, task=None):
        provider_args = nuka.config.get(self.provider.lower(), {})
        node_args = provider_args.get('create_node_args', {}).copy()
        node_args.update(self.vars.get('create_node_args'))
        self.log.debug4(nuka.utils.json.dumps(node_args, indent=2))
        return node_args

    async def destroy(self, **kwargs):
        raise NotImplementedError(
            'Please use Cloud.destroy() which is mush faster')


class GCEHost(Host):
    """Host on GCE"""

    use_sudo = True
    provider = Provider.GCE

    def get_create_node_args(self, driver=None, task=None):
        node_args = super().get_create_node_args(driver=driver, task=task)

        # update ex_metadata ssh keys
        keys_list = nuka.config.get('ssh', {}).get('keys') or []
        keysfile = nuka.config.get('ssh', {}).get('keysfile')
        if keysfile:
            with codecs.open(os.path.expanduser(keysfile), 'r', 'utf8') as fd:
                for line in fd:
                    keys_list.append(line.strip())
        if keys_list:
            node_args.setdefault('ex_metadata', {})
            items = node_args['ex_metadata'].setdefault('items', [])
            if not [v for v in items if v["key"] == "sshKeys"]:
                keys = []
                for line in keys_list:
                    if ':' in line:
                        keys.append(line)
                    else:
                        keys.append(self.vars['user'] + ':' + line)
                keys = '\n'.join(keys)
                items.append({"value": keys, "key": "sshKeys"})

        self.log.debug4(nuka.utils.json.dumps(node_args, indent=2))

        if 'ex_network' not in node_args:
            with self.timeit(type='api_call', task=task,
                             name='Driver.ex_list_networks()'):
                net = driver.ex_list_networks()[0]
            node_args['ex_network'] = net

        return node_args


class OpenstackHost(Host):
    """Host on OpenStack"""

    use_sudo = True
    provider = Provider.OPENSTACK

    def _set_ips(self):
        for iface in self.node.interface_list():
            for addr in iface.fixed_ips:
                ip = addr['ip_address']
                try:
                    a = ipaddress.IPv4Address(ip)
                except ValueError:
                    continue
                else:
                    if not a.is_private:
                        if 'public_ip' not in self.vars:
                            self.vars['public_ip'] = ip
                    else:
                        if 'private_ip' not in self.vars:
                            self.vars['private_ip'] = ip
        if 'private_ip' not in self.vars:
            # we may not have a private net
            self.vars['private_ip'] = self.vars['public_ip']

    @property
    def public_ip(self):
        """return host's public ip"""
        if 'public_ip' not in self.vars:
            self._set_ips()
        return self.vars['public_ip']

    @property
    def private_ip(self):
        if 'private_ip' not in self.vars:
            self._set_ips()
        return self.vars['private_ip']

    def get_create_node_args(self, driver=None, task=None):
        node_args = super().get_create_node_args(driver=driver, task=task)

        # update ex_metadata ssh keys
        keys_list = nuka.config.get('ssh', {}).get('keys') or []
        keysfile = nuka.config.get('ssh', {}).get('keysfile')
        if keysfile:
            with codecs.open(os.path.expanduser(keysfile), 'r', 'utf8') as fd:
                for line in fd:
                    keys_list.append(line.strip())
        if keys_list:
            driver = driver_from_config(self.provider, **self.driver_args)
            keypairs = [k for k in driver.keypairs.list() if k.name == 'ops']
            if not keypairs:
                driver.keypairs.create(
                    name='ops',
                    public_key='\n'.join(keys_list),
                )
            node_args['key_name'] = 'ops'
        self.log.debug4(nuka.utils.json.dumps(node_args, indent=2))
        if 'flavor' in node_args:
            node_args['flavor'] = driver.flavors.find(name=node_args['flavor'])
        if 'image' in node_args:
            node_args['image'] = driver.glance.find_image(node_args['image'])
        return node_args


class Cloud(base.HostGroup):
    """A HostGroup of cloud hosts"""

    _list_lock = threading.Lock()
    _nodes = {}

    host_classes = {
        Provider.GCE: GCEHost,
        Provider.OPENSTACK: OpenstackHost,
    }

    def __init__(self, provider=None, driver_args=None,
                 create=True, use_sudo=False):
        if provider is None:
            raise RuntimeError('A cloud instance require a valid provider')
        self.provider = provider
        self.driver_args = driver_args or {}
        self.create = create
        if provider in self.host_classes:
            self.host_class = self.host_classes[self.provider]
        else:
            name = '{0}Host'.format(self.provider.title())
            args = {'provider': self.provider, 'use_sudo': use_sudo}
            self.host_class = type(name, (Host,), args)
        self._nodes = {}

    @property
    def driver(self):
        return driver_from_config(self.provider, **self.driver_args)

    @property
    def cached_list_nodes(self):
        cls = self.__class__
        if not self._nodes:
            cls._list_lock.acquire()
            if self.provider in nova_providers:
                self._nodes = {n.name: n for n in self.driver.servers.list()}
            else:
                self._nodes = {n.name: n for n in self.driver.list_nodes()}
            cls._list_lock.release()
        return self._nodes

    def __getitem__(self, item):
        if item not in self:
            item = item.replace('_', '-')
        return super().__getitem__(item)

    def get_or_create_node(self, hostname, **kwargs):
        """Return a Host. Create it if needed"""
        kwargs.setdefault('create', self.create)
        kwargs['driver_args'] = self.driver_args
        if hostname not in self:
            kwargs.setdefault('node', self._nodes.get(hostname))
            self[hostname] = self.host_class(hostname=hostname, **kwargs)
        return self[hostname]

    def get_node(self, hostname, **kwargs):
        """Return a Host. Create it if needed"""
        kwargs['create'] = False
        host_node = self._nodes.get(hostname)
        return self.get_or_create_node(hostname, node=host_node, **kwargs)

    def from_compose(self, project_name=None, filename='docker-compose.yml'):
        """Return a host group with hosts names extracted from compose file"""
        from compose.project import Project
        from compose import config

        project_dir = os.getcwd()
        project_name = project_name or os.path.basename(project_dir)
        conf = config.load(
            config.find(project_dir, [filename], os.environ)
        )
        project = Project.from_config('', conf, client=None)
        for service in project.get_services():
            name = '{0}-{1.name}-1'.format(project_name, service)
            self.get_or_create_node(hostname=name)
        return self

    async def destroy(self):
        """Destroy all hosts in the group"""
        nodes = []
        for node in self.cached_list_nodes.values():
            if node.name in self:
                nodes.append(node)
                self[node.name].log.info('destroy()')
        if nodes:
            if self.provider in nova_providers:
                for node in nodes:
                    try:
                        node.force_delete()
                    except:
                        if node.name in self:
                            host = self[node.name]
                            host.log.exception('destroy()')
                        else:
                            raise

            else:
                self.driver.ex_destroy_multiple_nodes(nodes)
            del self[node.name]


def get_cloud(provider=Provider.GCE):
    """return a :class:`nuka.hosts.Cloud` filled with all instanciated hosts"""
    cloud = Cloud(provider=provider)
    for node in cloud.cached_list_nodes.values():
        cloud.get_or_create_node(node.name, node=node)
    return cloud



import sys
import argparse

from nuka.configuration import config


class Cli(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        kwargs['add_help'] = True
        super().__init__(*args, **kwargs)
        self.finalized = False
        self.args = None
        self.help = None
        self.add_argument(
            '-c', '--config', type=argparse.FileType('r'),
            required=False, default=None,
            help='yaml config file')
        self.add_argument('--diff', action='store_true', default=False,
                          help='run in diff mode')

        verbosity = self.add_argument_group('verbosity')
        verbosity.add_argument('-v', '--verbose', action='count', default=0,
                               help='increase verbosity')
        verbosity.add_argument('-q', '--quiet',
                               action='store_true', default=False,
                               help='log to stoud.log instead of stdout')
        verbosity.add_argument('--debug', action='store_true', default=False,
                               help='enable asyncio debug')

        dirs = self.add_argument_group('directories')
        dirs.add_argument(
            '--tempdir', default=None,
            help='tempdir name to store file localy and remotly')
        dirs.add_argument(
            '--nuka-dir', default=None,
            help='directory to store logs & reports. Default: .nuka')

        proc = self.add_argument_group('processes')
        proc.add_argument('-d', '--connections-delay', type=float,
                          metavar='DELAY', default=.2,
                          help='delay ssh connections. Default: 0.2')
        misc = self.add_argument_group('misc')
        misc.add_argument('--ssh', action='store_true', default=False,
                          help='use ssh binary instead of asyncssh')
        misc.add_argument('--uvloop', action='store_true', default=False,
                          help='use uvloop as eventloop')

    @property
    def arguments(self):
        return list(self._option_string_actions)

    def parse_known_args(self, *args, **kwargs):
        self.args, argv = super().parse_known_args(*args, **kwargs)
        return self.args, argv

    def print_help(self):
        self.help = True
        # Ignore warnings like:
        # sys:1: RuntimeWarning: coroutine 'do_something' was never awaited
        import warnings
        warnings.simplefilter("ignore")
        super().print_help()

    def parse_args(self, *args, **kwargs):
        self.args = super().parse_args(*args, **kwargs)
        if self.args.config:
            config.update_from_file(self.args.config)
        config.finalize(self.args)
        self.finalized = True


cli = Cli(add_help=False)

if config['testing']:
    cli.parse_args(['-vvvvvv', '--tempdir=/tmp/nuka_provisionning'])



# -*- coding: utf-8 -*-
"""
"""
import os

from nuka.task import Task
from nuka.tasks import http


class untar(Task):

    def __init__(self, src=None, dst=None, **kwargs):
        kwargs.setdefault('name', src)
        super(untar, self).__init__(src=src, dst=dst, **kwargs)

    def do(self):
        remove = False
        src = self.args['src']
        if src.startswith('http'):
            remove = True
            res = self.check(http.fetch(src=src).do())
            src = res['dst']
        res = self.sh(['tar', '-C', self.args['dst'], '-xzf', src])
        if remove:
            os.unlink(src)
        return res



#!../bin/python
# -*- coding: utf-8 -*-
import nuka
from nuka.hosts import Cloud
from nuka.tasks import shell

nuka.config['openstack'] = {
    'driver': '~/.ovh/nuka.json',
    'user': 'debian',
    'create_node_args': {
        'flavor': 's1-2',
        'image': 'Debian 8',
        'key_name': 'gawel',
    },
}

nuka.config['ssh'].update({
    'extra_options': ['-C', '-oStrictHostKeyChecking=no'],
    'keysfile': '~/.ssh/authorized_keys',
})


cloud = Cloud('openstack')
host = cloud.get_or_create_node('myhost')


async def my_tasks(host):
    await shell.command(['ls'])

nuka.cli.add_argument('--destroy', action='store_true', default=False)
nuka.cli.parse_args()

if nuka.cli.args.destroy:
    nuka.run(cloud.destroy())
else:
    nuka.run(my_tasks(host))



# Copyright 2017 by Bearstech <py@bearstech.com>
#
# This file is part of nuka.
#
# nuka is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nuka is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nuka. If not, see <http://www.gnu.org/licenses/>.

import time
import codecs
import inspect
import asyncio
import logging
import importlib

import asyncssh.misc

from nuka.remote.task import RemoteTask
from nuka.configuration import config
from nuka import remote
from nuka import utils
from nuka import gpg
import nuka


class Base(asyncio.Future):

    def __init__(self, **kwargs):
        self.initialize(**kwargs)
        super().__init__(loop=self.host.loop)
        if self.host.cancelled():
            self.cancel()
        else:
            self.process()

    def initialize(self, host=None,
                   switch_user=None, switch_ssh_user=None, **args):
        meta = {'filename': None, 'lineno': None,
                'start': time.time(), 'times': [],
                'remote_calls': [],
                }
        for infos in inspect.stack(2):
            f = infos.frame
            if isinstance(f.f_locals.get('self'), RemoteTask):
                continue
            if host is None:
                host = f.f_locals.get('host')
            if switch_user is None:
                switch_user = f.f_locals.get('switch_user')
            if switch_ssh_user is None:
                switch_ssh_user = f.f_locals.get('switch_ssh_user')
            if meta['filename'] is None:
                filename = infos.filename
                if filename.endswith('nuka/task.py'):
                    filename = 'nuka/task.py'
                meta.update(filename=filename,
                            lineno=infos.lineno)
                if host is not None:
                    break
        if host is None:  # pragma: no cover
            raise RuntimeError('No valid host found in the stack')
        self.switch_user = switch_user
        self.switch_ssh_user = switch_ssh_user
        self.meta = meta
        self.host = host
        self.loop = self.host.loop
        self.args = args
        self.res = {'changed': True, 'rc': 0}
        self.start = time.time()
        self.run_task = None
        host.add_task(self)

    def running(self):
        """return True if a remote task is running"""
        if self.run_task is not None:
            return not self.run_task.done()
        return False

    def process(self, fut=None):
        if fut is not None:  # pragma: no cover
            # we waited for boot
            self.meta['start'] = time.time()
        start = time.time()
        try:
            self.pre_process()
        except Exception as e:
            self.host.log.exception(e)
            self.cancel()
            raise
        else:
            duration = time.time() - start
            if duration > .05:  # pragma: no cover
                self.host.add_time(
                    start=start, time=duration,
                    type='pre_process', task=self)
            self.run_task = self._loop.create_task(self._run())

    def pre_process(self):
        """run locally before anything is sent to the host"""

    def post_process(self):
        """run when we get a valid reply from the host"""

    def render_template(self, fd):
        """render a template from a file descriptor:

        .. code-block:: python

            {'src': path, 'dst': path}
        """
        src = fd['src']
        ctx = dict(self.args, **self.args.get('ctx', {}))
        ctx.update(host=self.host, env=config, **fd)
        engine = config.get_template_engine()
        template = engine.get_template(src)
        fd['data'] = template.render(ctx)
        if 'executable' not in fd:
            fd['executable'] = utils.isexecutable(src)

    def render_file(self, fd):
        """render a file from a file descriptor. A file descriptor is a dict:

        .. code-block:: python

            {'src': path, 'dst': path}
        """
        src = fd['src']
        if src.endswith('.gpg'):
            _, data = gpg.decrypt(src, 'utf8')
        else:
            with codecs.open(src, 'r', 'utf8') as fd_:
                data = fd_.read()
        fd['data'] = data
        if 'executable' not in fd:
            fd['executable'] = utils.isexecutable(src)

    def log(self):
        self.host.log.info(self)

    def cancel(self):
        """cancel a task"""
        if not self.cancelled():
            super().cancel()
            if not self.res.get('signal') and not self.host.failed():
                # do not log cancellation if the user wanted it
                self.log()
            if self.run_task is not None:
                self.run_task.cancel()
            self.host.cancel()

    async def _run(self):
        # wrap the task to catch exception
        try:
            await self.run()
        except Exception as e:
            self.cancel()
            if not isinstance(e, asyncio.CancelledError):
                self.host.log.exception5(self)

        # update meta
        self.meta.update(self.res.pop('meta', {}))

        # if task succeded then run post_process
        start = time.time()
        try:
            self.post_process()
        except Exception:
            self.cancel()
            self.host.log.exception5(self)
        finally:
            duration = time.time() - start
            if duration > .05:
                self.host.add_time(
                    start=start, time=duration,
                    type='post_process', task=self)

        # set result / log stuff
        if not self.done():
            self.set_result(self)
            self.meta.setdefault('time', time.time() - self.meta['start'])
            self.host.add_time(type='task', task=self, **self.meta)

        # log if not cancelled
        if not self.cancelled():
            self.log()

    def __bool__(self):
        return self.res.get('rc') == 0

    def __repr__(self):
        return '<{0}>'.format(str(self))

    def __str__(self):
        name = self.__class_name__()
        instance_name = self.args.get('name')
        if instance_name is None:
            instance_name = '-'
        s = '{0}({1})'.format(name, instance_name)
        if self.res:
            if self.res['rc'] == 0:
                if self.cancelled():
                    s += ' cancelled at {filename}:{lineno}'.format(
                        **self.meta)
                elif self.done() and getattr(self, 'changed', True):
                    s += ' changed'
            else:
                s += ' fail({0[rc]})'.format(self.res)
            time = self.meta.get('local_time')
            if time:
                s += ' time({0}s)'.format(round(time, 1))
        return s.strip()

    def __class_name__(self):
        klass = self.__class__
        name = '{0}.{1}'.format(klass.__module__.split('.')[-1],
                                klass.__name__)
        return name


class Task(Base, RemoteTask):

    def process(self):
        if self.host.cancelled():
            self.cancel()
        else:
            diff_mode = self.args.get('diff_mode', nuka.cli.args.diff)
            if diff_mode:
                # ignore diff call if the task do not support it
                attr = getattr(self, 'diff', None)
                if attr in (None, False):
                    self.res['changed'] = False
                    self.meta['local_time'] = 0.
                    if attr is False:
                        self.host.log.info("{0}.diff is False".format(self))
                    else:
                        self.host.log.warning("{0}.diff is None".format(self))
                    self.set_result(self)
                    return
            if self.host.fully_booted.done():
                super().process()
            else:
                # use asyncio with callback since we are in a sync __init__
                task = self.loop.create_task(wait_for_boot(self.host))
                task.add_done_callback(super().process)

    async def run(self):
        """Serialize the task, send it to the remote host.
        The remote script will deserialize the task and run
        :meth:`~nuka.remote.task.Task.do` (or diff() when using --diff)
        """
        self.host.log.debug(self)
        diff_mode = self.args.get('diff_mode', nuka.cli.args.diff)
        klass = self.__class__

        args = {}
        for k, v in self.args.items():
            if k not in ('ctx',):
                args[k] = v

        # prep stdin
        stdin_data = dict(
            task=(klass.__module__, klass.__name__),
            remote_tmp=config['remote_tmp'],
            switch_user=self.switch_user,
            args=args,
            check_mode=False,
            diff_mode=diff_mode,
            log_level=config['log']['levels']['remote_level'])

        if config['testing'] and 'coverage' in self.host.vars:
            # check if we can/want use coverage
            cmd = (
                '{coverage} run -p '
                '--source={remote_dir}/nuka/tasks '
                '{script} '
            ).format(coverage=self.host.vars['coverage'], **config)
        else:
            # use python
            inventory = self.host.vars.get(
                'inventory',
                {'python': {'executable': 'python'}})
            executable = inventory['python'].get('executable', 'python')
            cmd = '{0} {script} '.format(executable, **config)

        # allow to trac some ids from ps
        cmd += '--deploy-id={0} --task-id={1}'.format(config['id'],
                                                      id(self))
        # create process
        proc = await self.host.create_process(
            cmd, task=self,
            switch_user=self.switch_user,
            switch_ssh_user=self.switch_ssh_user)

        # send stdin
        zlib_avalaible = self.host.inventory['python']['zlib_available']
        stdin = utils.proto_dumps_std(
            stdin_data, proc.stdin,
            content_type=zlib_avalaible and 'zlib' or 'plain')
        await proc.stdin.drain()

        res = {}
        while res.get('message_type') != 'exit':
            # wait for messages
            try:
                res = await proc.next_message()
            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.cancel()
                self.host.log.exception5(
                    '{0}\n\n{1}'.format(self, stdin))
            else:
                if res.get('message_type') == 'log':
                    self.host.log.log(res['level'], res['msg'])

        # finalize
        self.res.update(res)
        if self.res['rc'] != 0 and not self.ignore_errors:
            if not diff_mode:
                self.cancel()

    def log(self):
        log = self.host.log
        if 'exc' in self.res:
            exc = '\n' + ''.join(self.res['exc'])
            log.error('{0}\n{1}'.format(self, exc))
        elif self.res.get('stderr'):
            if self.res.get('rc') != 0:
                log.error('{0}\n{1}'.format(self, self.res['stderr']))
            elif self.res['changed']:
                log.changed('{0}\n{1}'.format(self, self.res['stderr']))
        else:
            data = self.res.get('diff', '')
            if data.strip():
                data = data.strip()
                log.changed('{0} diff=\n{1}\n'.format(self, data))
            elif self.cancelled():
                log.error(self)
            elif self.res['changed']:
                log.changed(self)
            else:
                log.info(self)

        for cmd_ in self.meta.get('remote_calls', []):
            rtime = round(cmd_['time'], 3)
            inf = '^ sh({cmd}) time({rtime})'.format(
                rtime=rtime, **cmd_)
            if cmd_['exc']:
                log.error(inf + '\n' + ''.join(cmd_['exc']))
            elif nuka.cli.args.verbose > 1:
                log.info(inf)
                stds = {k: v for k, v in cmd_.items()
                        if k in ('stderr', 'stdout') and v}
                if stds:
                    log.debug3('^ ' + str(stds))

        data = self.res.get('log')
        if data:
            level = None
            for line in data.rstrip().split('\n'):
                line = line.rstrip()
                try:
                    line_level, message = line.split(':', 1)
                except ValueError:
                    if level:
                        log.log(level, '^ ' + line)
                else:
                    if line_level in logging._nameToLevel:
                        level = getattr(logging, line_level)
                        log.log(level, '^ ' + message)
                    else:
                        if level:
                            log.log(level, '^ ' + line)


class SetupTask(Base):

    changed = False

    def log(self):
        self.host.log.debug(self)

    def pre_process(self):
        self.args['name'] = ''  # unamed task

    def cancel(self):
        super().cancel()
        if not self.host.cancelled():
            self.host.cancel()
            self.host.log.critical(
                'Cancelled at {filename}:{lineno}...'.format(**self.meta))


class boot(SetupTask):
    """A task that just call host.boot()"""

    def __class_name__(self):
        return 'boot'

    async def run(self):
        # wait for boot async
        try:
            await self.host.boot()
        except:
            self.host.log.exception('boot')
        self.meta['start'] = self.host._start


class setup(SetupTask):
    """A task that just wait for :class:`~nuka.task.boot` then setup the
    host"""

    setup_cmd = (
        '{0}rm -Rf {2[remote_tmp]}; '
        '{0}mkdir -p {2[remote_tmp]} && {0}chmod 777 {2[remote_tmp]} &&'
        '{0}mkdir -p {2[remote_dir]} &&'
        'dd bs={1} count=1 | {0}tar -xz -C {2[remote_dir]} && '
        '{0}`which python 2> /dev/null || which python3 || echo python` '
        '{2[script]} --setup'
    )

    def __class_name__(self):
        return 'setup'

    async def run(self):
        host = self.host
        # wait for boot async
        await host._named_tasks[boot.__name__]
        self.meta['start'] = time.time()

        # run bootstrap_command if any
        if host.bootstrap_command:
            res = await host.run_command(host.bootstrap_command)
            if res['rc'] != 0:
                self.host.log.error(res)
                self.cancel()
                return

        # setup
        sudo = ''
        if host.use_sudo:
            sudo = '{sudo} '.format(**config)

        cmd = self.setup_cmd.format(sudo, '{bytes}', config)

        mods = nuka.config['inventory_modules'][:]
        mods += self.host.vars.get('inventory_modules', [])
        if mods:
            cmd += ' ' + ' '.join(['--inventory=' + m for m in mods])

        stdin = remote.build_archive(
            extra_classes=all_task_classes(),
            mode='x:gz')
        c = cmd.format(bytes=len(stdin))
        host.log.debug('Uploading archive ({0}kb)...'.format(
                int(len(stdin) / 1000)))

        try:
            proc = await self.host.create_process(c, task=self)
            proc.stdin.write(stdin)
            await proc.stdin.drain()
        except (LookupError, OSError, asyncssh.misc.Error) as e:
            if isinstance(e, asyncssh.misc.Error):
                e = LookupError(str(e), self.host)
            self.host.log.error(e.args[0])
            self.host.fail(e)
            return

        res = {}
        while res.get('message_type') != 'exit':
            # wait for messages
            try:
                res = await proc.next_message()
            except asyncio.CancelledError:
                raise
            except (LookupError, OSError) as e:
                self.host.log.error(e.args[0])
                self.host.fail(e)
                return
            except Exception as e:
                self.cancel()
                self.host.log.exception5(
                    '{0}\n\n{1}'.format(self, stdin))
            else:
                if res.get('message_type') == 'log':
                    self.host.log.log(res['level'], res['msg'])

        self.res.update(res)

        if self.res['rc'] != 0:
            self.cancel()

        host.vars['inventory'] = self.res['inventory']
        for name in mods:
            mod = importlib.import_module(name)
            meth = getattr(mod, 'finalize_inventory', None)
            if meth is not None:
                meth(host.vars['inventory'])
        host.log.debug(
            'Inventory:\n{0}'.format(host.vars['inventory']))

        if not host.fully_booted.done():
            host.fully_booted.set_result(True)


class teardown(SetupTask):
    """remove `remote_dir` from the host"""

    teardown_cmd = '{0}rm -Rf {1[remote_dir]}'

    def __init__(self, host):
        host._cancelled = False
        super().__init__(host=host)

    def __class_name__(self):
        return 'teardown'

    async def run(self):
        if not self.host.failed():
            sudo = self.host.use_sudo and 'sudo ' or ''
            cmd = self.teardown_cmd.format(sudo, config)
            await self.host.run_command(cmd, task=self)


class destroy(SetupTask):
    """destroy the host"""

    def __init__(self, host):
        host._cancelled = False
        super().__init__(host=host)

    def __class_name__(self):
        return 'destroy'

    async def run(self):
        if 'destroyed' not in self.host.vars:
            await self.host.destroy()


class wait(Base):
    """A task that wait for a coroutine / event / future:

    .. code-block:: python

        nuka.wait(do_something(host), event)

    You can use a timeout:

    .. code-block:: python

        nuka.wait(event, timeout=30)

    """

    def __class_name__(self):
        return 'wait'

    def __init__(self, future, *futures, **kwargs):
        futures = list(futures)
        if not isinstance(future, list):
            futures.insert(0, future)
        else:
            futures[0:0] = future
        kwargs['name'] = repr(futures)
        kwargs['futures'] = futures
        super().__init__(**kwargs)

    async def run(self):
        futures = self.args['futures']

        res = await asyncio.wait_for(asyncio.gather(*futures),
                                     timeout=self.args.get('timeout'))

        self.set_result(res)
        self.meta.setdefault('time', time.time() - self.meta['start'])

        # skip time if we dont wait for a nuka.Event
        events = [e for e in res if isinstance(e, nuka.Event)]
        if events:
            self.host.add_time(type='task', task=self, **self.meta)


async def wait_for_boot(host):
    if not host.fully_booted.done():
        create_setup_tasks(host)
        task = host._named_tasks[setup.__name__]
        if not task.done():
            await task


def create_setup_tasks(host):
    if not host.fully_booted.done():
        for task in (boot, setup):
            instance = host._named_tasks.get(task.__name__)
            if instance is None:
                instance = task(host=host)
                host._named_tasks[task.__name__] = instance
                host._tasks.append(instance)


def get_task_from_stack():
    for info in inspect.stack(3):
        f = info.frame
        self = f.f_locals.get('self')
        if isinstance(f.f_locals.get('self'), Base):
            return self


def all_task_classes(cls=Task):
    for klass in cls.__subclasses__():
        yield from all_task_classes(klass)
        yield klass



# -*- coding: utf-8 -*-
"""
file related tasks
"""
from nuka.task import Task
from nuka import utils
import logging
import codecs
import stat
import glob
import re
import os


class exists(Task):
    """return True if a path exists"""

    ignore_errors = True

    def __init__(self, dst=None, **kwargs):
        kwargs.setdefault('name', dst)
        kwargs.setdefault('dst', dst)
        super(exists, self).__init__(**kwargs)

    def do(self):
        dst = self.args['dst']
        if os.path.exists(dst):
            return dict(rc=0, dst=dst, changed=False)
        return dict(rc=1, dst=dst)

    def diff(self):
        dst = self.args['dst']
        if os.path.exists(dst):
            return dict(rc=0, dst=dst, diff='', changed=False)
        diff = self.texts_diff('', dst)
        return dict(rc=1, dst=dst, diff=diff)


class mkdir(Task):
    """create a directory"""

    def __init__(self, dst=None, mod=None, own=None, **kwargs):
        kwargs.setdefault('name', dst)
        kwargs.update(dst=dst, mod=mod, own=None)
        super(mkdir, self).__init__(**kwargs)

    def do(self):
        dst = self.args['dst']
        mod = self.args['mod']
        own = self.args['mod']
        res = utils.makedirs(dst, mod=mod, own=own)
        res.update(rc=0, dst=dst)
        return res

    def diff(self):
        dst = self.args['dst']
        diff = ''
        if not os.path.exists(dst):
            res = self.lists_diff([], [dst + '\n'], fromfile=dst)
            diff = u''.join(res) + u'\n'
        return dict(rc=0, diff=diff, dst=dst)


class mkdirs(Task):
    """create directories"""

    def __init__(self, directories=None, **kwargs):
        kwargs.setdefault('name', ', '.join([d['dst'] for d in directories]))
        super(mkdirs, self).__init__(directories=directories, **kwargs)

    def do(self):
        res = dict(rc=0, changed=[])
        for d in self.args['directories']:
            dst = d.pop('dst')
            r = utils.makedirs(dst, **d)
            if r['changed']:
                res['changed'].append(dst)
        return res

    def diff(self):
        res = dict(rc=0, diff='', changed=[])
        for d in self.args['directories']:
            dst = d.pop('dst')
            r = mkdir(dst, **d).diff()
            if r['diff']:
                res['diff'] += r['diff']
                res['changed'].append(dst)
        return res


class rm(Task):
    """rm a file or directory"""

    def __init__(self, dst=None, **kwargs):
        kwargs.setdefault('name', dst)
        kwargs.setdefault('dst', dst)
        super(rm, self).__init__(**kwargs)

    def do(self):
        changed = False
        for dst in glob.glob(self.args['dst']):
            if os.path.exists(dst):
                try:
                    if os.path.isdir(dst):
                        import shutil
                        shutil.rmtree(dst)
                    else:
                        os.remove(dst)
                    changed = True
                except OSError:
                    exc = self.format_exception()
                    return dict(rc=1, exc=exc)
        return dict(rc=0, changed=changed)

    def diff(self):
        diff = ''
        filenames = []
        for dst in glob.glob(self.args['dst']):
            if os.path.exists(dst):
                filenames.append(dst + '\n')
        if filenames:
            res = self.lists_diff(filenames, [], fromfile=dst)
            diff = u''.join(res) + u'\n'
        return dict(rc=0, diff=diff)


class mv(Task):
    """rename/move a file or a directory"""

    def __init__(self, src=None, dst=None, **kwargs):
        kwargs.setdefault('name', src)
        super(mv, self).__init__(src=src, dst=dst, **kwargs)

    def do(self):
        src = self.args['src']
        dst = self.args['dst']
        if not os.path.exists(src) and os.path.exists(dst):
            # Consider the task already done
            return dict(rc=0, changed=False)
        try:
            os.rename(src, dst)
        except OSError:
            exc = self.format_exception()
            return dict(rc=1, exc=exc)
        return dict(rc=0, changed=True)

    def diff(self):
        src = self.args['src']
        dst = self.args['dst']
        diff = ''
        if os.path.exists(src):
            diff += "%s -> %s\n" % (src, dst)
        return dict(rc=0, diff=diff)


class chmod(Task):
    """apply chmod to dst"""

    diff = False

    def __init__(self, dst=None, mod='644', recursive=False, **kwargs):
        kwargs.setdefault('name', dst)
        super(chmod, self).__init__(dst=dst, mod=mod,
                                    recursive=recursive, **kwargs)

    def do(self):
        dst = self.args['dst']
        mod = self.args['mod']
        recursive = self.args['recursive']
        utils.chmod(dst, mod, recursive=recursive)
        return dict(rc=0, dst=dst, mod=mod)


class chown(Task):
    """apply chown to dst"""

    diff = False

    def __init__(self, dst=None, own='root:root', recursive=False, **kwargs):
        kwargs.setdefault('name', dst)
        super(chown, self).__init__(dst=dst, own=own,
                                    recursive=recursive, **kwargs)

    def do(self):
        dst = self.args['dst']
        own = self.args['own']
        recursive = self.args['recursive']
        utils.chown(dst, own, recursive=recursive)
        return dict(rc=0, dst=dst, own=own)


class put(Task):
    """put files on the remote host"""

    def __init__(self, files=None, **kwargs):
        kwargs.setdefault('name', [f['dst'] for f in files or []])
        super(put, self).__init__(files=files, **kwargs)

    def pre_process(self):
        for fd in self.args['files']:
            if 'linkto' in fd:
                fd['data'] = None
            elif 'src' in fd:
                src = fd['src']
                if src.startswith('~/'):
                    fd['src'] = os.path.expanduser(src)
                if src.endswith(('.j2', '.j2.gpg')):
                    self.render_template(fd)
                else:
                    self.render_file(fd)
            elif 'tpl' in fd:
                src = fd['src'] = fd['tpl']
                if src.startswith('~/'):
                    fd['src'] = os.path.expanduser(src)
                self.render_template(fd)
            if 'data' not in fd:
                raise RuntimeError('cant get content for fd {0}'.format(fd))

    def do(self):
        files = self.args['files'] or []
        files_changed = []
        for fd in files:
            dst = fd['dst']
            if dst.startswith('~/'):
                dst = fd['dst'] = os.path.expanduser(dst)
            if 'linkto' in fd:
                link = fd['linkto']
                if os.path.exists(dst):
                    if os.path.islink(dst):
                        if os.path.realpath(dst) != os.path.realpath(link):
                            os.unlink(dst)
                    else:
                        os.remove(dst)
                if not os.path.exists(link):
                    self.send_log((
                        'Invalid link destination for '
                        '{0[dst]} -> no such file {0[linkto]}'
                        ).format(fd), level=logging.ERROR)
                if not os.path.islink(dst):
                    os.symlink(link, dst)
                    files_changed.append(dst)
            else:
                data = fd['data']
                if os.path.exists(dst):
                    with codecs.open(dst, 'rb', 'utf8') as fd_:
                        if data != fd_.read():
                            files_changed.append(dst)
                else:
                    files_changed.append(dst)
                with codecs.open(dst, 'wb', 'utf8') as fd_:
                    fd_.write(data)
            mod = fd.get('mod')
            if mod is None and fd.get('executable'):
                st = os.stat(dst)
                mod = (
                    st.st_mode | stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH
                )
            if mod is not None:
                utils.chmod(dst, mod)
        return dict(rc=0, changed=files_changed)

    def diff(self):
        diff = ''
        files = self.args['files'] or []
        files_changed = []
        for fd in files:
            dst = fd['dst']
            if 'linkto' in fd:
                if os.path.exists(fd['linkto']):
                    new_text = '{0[dst]} -> {0[linkto]}\n'.format(fd)
                else:
                    new_text = (
                        '{0[dst]} -> no such file {0[linkto]}\n'
                    ).format(fd)
                    self.send_log((
                        'Invalid link destination for {0}'
                        ).format(new_text.strip()),
                        level=logging.ERROR)
                if not os.path.isfile(dst):
                    old_text = ''
                else:
                    old_text = new_text
            else:
                new_text = fd['data']
                if not os.path.isfile(dst):
                    old_text = ''
                else:
                    with codecs.open(dst, 'rb', 'utf8') as fd_:
                        old_text = fd_.read()
            if old_text != new_text:
                files_changed.append(dst)
            diff += self.texts_diff(old_text, new_text, fromfile=dst)
        return dict(rc=0, diff=diff, changed=files_changed)


class scripts(put):
    """put and execute scripts on the remote host"""

    def do(self):
        files = self.args['files'] or []
        for f in files:
            f['executable'] = True
        super(scripts, self).do()
        for f in files:
            res = self.sh(f['dst'])
        return res


class cat(Task):
    """cat a file"""

    def __init__(self, src=None, **kwargs):
        kwargs.setdefault('name', src)
        super(cat, self).__init__(src=src, **kwargs)

    def do(self):
        src = self.args['src']
        with codecs.open(src, 'rb', 'utf8') as fd:
            res = dict(rc=0, content=fd.read())
        return res


class update(Task):
    """update a file"""

    def __init__(self, dst=None, replaces=None, appends=None, **kwargs):
        kwargs.setdefault('name', dst)
        super(update, self).__init__(
            dst=dst, replaces=replaces, appends=appends, **kwargs)

    def pre_process(self):
        for regex, value in self.args['replaces']:
            re.compile(regex)

    def update(self, data):
        for regex, value in self.args['replaces'] or []:
            data = re.sub(regex, value, data, flags=re.MULTILINE)
        lines = data.splitlines(1)
        for after, value in self.args['appends'] or []:
            value += '\n'
            if value not in lines:
                if after == 'EOF':
                    lines.append(value)
                else:
                    for i, line in enumerate(lines):
                        if re.match(after, line):
                            lines[i + 1:i + 1] = [value]
        return ''.join(lines)

    def do(self):
        dst = self.args['dst']
        with codecs.open(dst, 'r', 'utf8') as fd:
            old_data = data = fd.read()
        data = self.update(data)
        changed = old_data != data
        if changed:
            with codecs.open(dst, 'w', 'utf8') as fd:
                fd.write(data)
        return dict(rc=0, changed=changed)

    def diff(self, dst=None, **kwargs):
        dst = self.args['dst']
        with codecs.open(dst, 'r', 'utf8') as fd:
            old_data = data = fd.read()
        data = self.update(data)
        if old_data != data:
            diff = self.texts_diff(old_data, data, fromfile=dst)
        else:
            diff = u''
        return dict(rc=0, diff=diff)



# -*- coding: utf-8 -*-
"""
mysql related tasks
"""
import os
import codecs
import getpass

from nuka import utils
from nuka.task import Task


class my_cnf(Task):
    """create ~/.my.cnf"""

    def __init__(self, password=None, switch_user='root', **kwargs):
        kwargs['name'] = '~{0}/.my.cnf'.format(switch_user)
        super(my_cnf, self).__init__(switch_user=switch_user,
                                     password=password, **kwargs)

    def do(self):
        self.args['user'] = getpass.getuser()
        dst = os.path.expanduser('~/.my.cnf')
        old_data = ''
        if os.path.isfile(dst):
            with codecs.open(dst, 'r', 'utf8') as fd:
                old_data = fd.read()
        data = (
            '[client]\n'
            'user={user}\n'
            'password={password}\n').format(**self.args)
        changed = old_data != data
        if changed:
            with codecs.open(dst, 'w', 'utf8') as fd:
                fd.write(data)
            utils.chmod(dst, '600')
        self.args.pop('password')
        return dict(rc=0, changed=changed, dst=dst)

    def diff(self):
        self.args['user'] = getpass.getuser()
        dst = os.path.expanduser('~/.my.cnf')
        old_data = ''
        if os.path.isfile(dst):
            with codecs.open(dst, 'r', 'utf8') as fd:
                old_data = fd.read()
        data = (
            '[client]\n'
            'user={user}\n'
            'password={password}\n').format(**self.args)
        diff = self.texts_diff(old_data, data, fromfile=dst)
        return dict(rc=0, diff=diff)


class set_root_password(my_cnf):
    """set mysql root password and store it in ~/.my.cnf"""

    def __init__(self, password=None, **kwargs):
        kwargs.update(password=password, switch_user='root')
        super(set_root_password, self).__init__(**kwargs)

    def do(self):
        new_password = self.args['password']
        old_password = None

        # get old password
        dst = os.path.expanduser('/root/.my.cnf')
        if os.path.isfile(dst):
            with codecs.open(dst, 'r', 'utf8') as fd:
                for line in fd:
                    line = line.strip()
                    if line.startswith('password'):
                        old_password = line.split('=')[1].strip()

        if old_password != new_password:
            # change password
            cmd = ['mysqladmin', 'password', new_password]
            if old_password:
                cmd[1:1] = ['--password=' + old_password]
            self.sh(cmd)
        # forward to my_cnf to update /root/.my.cnf
        return super(set_root_password, self).do()


class create_db(Task):
    """create a database and grant user"""

    statement = '''
    CREATE DATABASE IF NOT EXISTS {name};
    GRANT ALL PRIVILEGES ON *.* TO '{user}'@'%' IDENTIFIED BY '{password}';
    FLUSH PRIVILEGES;
    '''

    def __init__(self, name=None, user=None, password=None, **kwargs):
        super(create_db, self).__init__(name=name, user=user,
                                        password=password, **kwargs)

    def do(self):
        return self.sh('mysql', stdin=self.statement.format(**self.args))


class execute(Task):
    """execute a sql statement"""

    diff = False

    def __init__(self, sql=None, **kwargs):
        kwargs.setdefault('name', sql)
        super(execute, self).__init__(sql=sql, **kwargs)

    def do(self):
        return self.sh('mysql', stdin=self.args['sql'])
